import os
import re
import sys
import timeit
from pathlib import PurePosixPath

import h5py
import numpy as np
import psutil
from mpi4py import MPI
from tqdm import tqdm

#from .rgd import rgd ## no!
from .gradient import gradient
from .h5 import h5_chunk_sizer, h5_print_contents
from .utils import even_print, format_time_string

# ======================================================================

def _calc_turb_budget_xpln(self, **kwargs):
    '''
    calculate budget of turbulent kinetic energy (k)
    -----
    - dimensional [SI]
    - designed for analyzing unsteady, thin planes in [x]
    -----
    SI units of terms are [kg/(m·s³)] or [kg m^-1 s^-3]
    normalize with
    * ν / u**4 / ρ --> *[ kg^-1 m s^3]
    or
    / ( u**4 * ρ / ν ) --> /[kg m^-1 s^-3]
    -----
    Pirozzoli Grasso Gatski (2004)
    Direct numerical simulation and analysis of a spatially evolving supersonic turbulent boundary layer at M=2.25
    https://doi.org/10.1063/1.1637604
    -----
    Gaurini Moser Shariff Wray (2000)
    Direct numerical simulation of a supersonic turbulent boundary layer at Mach 2.5
    https://doi.org/10.1017/S0022112000008466
    '''
    
    rgd_meta = type(self) ## workaround for using rgd()
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'rgd.calc_turb_budget_xpln()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if (self.fsubtype!='unsteady'):
        raise RuntimeError(f"file subtype must be 'unsteady' but is '{str(self.fsubtype)}'")
    
    if not self.usingmpi:
        raise ValueError('rgd.calc_turb_budget_xpln() currently only works in MPI mode')
    
    rx = kwargs.get('rx',1)
    ry = kwargs.get('ry',1)
    rz = kwargs.get('rz',1)
    rt = kwargs.get('rt',1)
    
    ## number of subdivisions per rank [t] range
    ct = kwargs.get( 'ct' , int((self.nt//self.n_ranks)//4) )
    if not isinstance(ct,int) or (ct<1):
        raise TypeError('ct should be a positive non-zero int')
    
    # st = kwargs.get('st',1)
    # if not isinstance(st,int) or (st<1):
    #     raise TypeError('st should be a positive non-zero int')
    # if (self.nt%st!=0):
    #     raise ValueError('nt not divisible by st')
    
    force = kwargs.get('force',False)
    
    fn_h5_mean = kwargs.get('fn_h5_mean',None)
    fn_h5_out  = kwargs.get('fn_h5_out',None)
    
    save_unsteady      = kwargs.get('save_unsteady',False)
    fn_h5_out_unsteady = kwargs.get('fn_h5_out_unsteady',None)
    
    acc          = kwargs.get('acc', 6)
    edge_stencil = kwargs.get('edge_stencil', 'half')
    
    chunk_kb         = kwargs.get('chunk_kb',4*1024) ## h5 chunk size: default 4 [MB]
    chunk_constraint = kwargs.get('chunk_constraint',(1,None,None,None)) ## the 'constraint' parameter for sizing h5 chunks
    chunk_base       = kwargs.get('chunk_base',2)
    
    ## only distribute data in [t]
    if (rx!=1):
        raise AssertionError('rx!=1')
    if (ry!=1):
        raise AssertionError('ry!=1')
    if (rz!=1):
        raise AssertionError('rz!=1')
    
    if not isinstance(rt,int) or (rt<1):
        raise ValueError('rt should be a positive non-zero int')
    
    if (rx*ry*rz*rt != self.n_ranks):
        raise AssertionError('rx*ry*rz*rt != self.n_ranks')
    if (rx>self.nx):
        raise AssertionError('rx>self.nx')
    if (ry>self.ny):
        raise AssertionError('ry>self.ny')
    if (rz>self.nz):
        raise AssertionError('rz>self.nz')
    if (rt>self.nt):
        raise AssertionError('rt>self.nt')
    
    # ===
    
    rtl_ = np.array_split(np.arange(self.nt,dtype=np.int64),rt)
    rtl = [[b[0],b[-1]+1] for b in rtl_ ]
    rt1,rt2 = rtl[self.rank]
    ntr = rt2 - rt1
    
    if (ntr<1):
        print(f'rank {self.rank:d} ntr < 1')
        self.comm.Abort(1)
    if (ntr<ct):
        print(f'rank {self.rank:d} ntr < ct')
        self.comm.Abort(1)
    
    ## [t] sub chunk range --> ctl = list of ranges in rt1:rt2
    ctl_ = np.array_split( np.arange(rt1,rt2) , ct )
    ctl = [[b[0],b[-1]+1] for b in ctl_ ]
    
    ## check that no sub ranges are <=1
    for a_ in [ ctl_[1]-ctl_[0] for ctl_ in ctl ]:
        if (a_ <= 1):
            #raise ValueError
            print(f'rank {self.rank:d} has [t] a chunk subrange <= 1')
            self.comm.Abort(1)
    
    ## the average sub [t] chunk size on this rank
    #avg_ntc = np.mean( [ ctl_[1]-ctl_[0] for ctl_ in ctl ] )
    #min_ntc =     min( [ ctl_[1]-ctl_[0] for ctl_ in ctl ] )
    #max_ntc =     max( [ ctl_[1]-ctl_[0] for ctl_ in ctl ] )
    
    ## get all ntcs
    my_ntcs = [ ctl_[1]-ctl_[0] for ctl_ in ctl ]
    G = self.comm.gather([ self.rank , my_ntcs ], root=0)
    G = self.comm.bcast(G, root=0)
    
    allntcs = []
    for G_ in G:
        allntcs += G_[1]
    allntcs = np.array( allntcs , dtype=np.int64 )
    
    avg_ntc = np.mean( allntcs , dtype=np.float64 )
    min_ntc = allntcs.min()
    max_ntc = allntcs.max()
    
    if 1: ## check that [t] sub-chunk ranges are correct
        
        mytimeindices = []
        for ctl_ in ctl:
            ct1, ct2 = ctl_
            mytimeindices += [ ti_ for ti_ in self.ti[ct1:ct2] ]
        
        G = self.comm.gather([ self.rank , mytimeindices ], root=0)
        G = self.comm.bcast(G, root=0)
        
        alltimeindices = []
        for G_ in G:
            alltimeindices += G_[1]
        alltimeindices = np.array( sorted(alltimeindices), dtype=np.int64 )
        
        if not np.array_equal( alltimeindices , self.ti ):
            raise AssertionError
        if not np.array_equal( alltimeindices , np.arange(self.nt, dtype=np.int64) ):
            raise AssertionError
    
    self.comm.Barrier()
    
    ## mean file name (for reading)
    if (fn_h5_mean is None):
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        fname_mean_h5_base = fname_root+'_mean.h5'
        #fn_h5_mean = os.path.join(fname_path, fname_mean_h5_base)
        fn_h5_mean = str(PurePosixPath(fname_path, fname_mean_h5_base))
        #fn_h5_mean = Path(fname_path, fname_mean_h5_base)
    
    if not os.path.isfile(fn_h5_mean):
        raise FileNotFoundError('%s not found!'%fn_h5_mean)
    
    ## turb_budget .h5 file name (for writing) --> AVERAGED
    if (fn_h5_out is None):
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
        fn_h5_out_base = fname_root+'_turb_budget.h5'
        fn_h5_out = str(PurePosixPath(fname_path, fn_h5_out_base))
    
    if os.path.isfile(fn_h5_out) and (force is False):
        raise ValueError(f'{fn_h5_out} already present & force=False')
    
    ## turb_budget .h5 file name (for writing) --> UNSTEADY
    if save_unsteady:
        if (fn_h5_out_unsteady is None):
            fname_path = os.path.dirname(self.fname)
            fname_base = os.path.basename(self.fname)
            fname_root, fname_ext = os.path.splitext(fname_base)
            fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
            fn_h5_out_unsteady_base = fname_root+'_turb_budget_unsteady.h5'
            fn_h5_out_unsteady = str(PurePosixPath(fname_path, fn_h5_out_unsteady_base))
        
        if os.path.isfile(fn_h5_out_unsteady) and (force is False):
            raise ValueError(f'{fn_h5_out_unsteady} already present & force=False')
    
    if verbose: even_print('fn_h5'      , self.fname )
    if verbose: even_print('fn_h5_mean' , fn_h5_mean )
    if verbose: even_print('fn_h5_out'  , fn_h5_out  )
    if verbose: print(72*'-')
    if verbose: even_print('save unsteady' , str(save_unsteady) )
    if verbose and save_unsteady:
        even_print('unsteady file' , fn_h5_out_unsteady )
    if verbose: print(72*'-')
    
    # ===
    
    if verbose: even_print('nx' , f'{self.nx:d}' )
    if verbose: even_print('ny' , f'{self.ny:d}' )
    if verbose: even_print('nz' , f'{self.nz:d}' )
    if verbose: even_print('nt' , f'{self.nt:d}' )
    if verbose: print(72*'-')
    if verbose: even_print('rt'  , f'{rt:d}' )
    if verbose: even_print('ct'  , f'{ct:d}' )
    #if verbose: even_print('st'  , f'{st:d}' )
    if verbose: even_print('ntr' , f'{ntr:d}' )
    if verbose: even_print('avg [t] chunk nt' , f'{avg_ntc:0.6f}' )
    if verbose: even_print('min [t] chunk nt' , f'{min_ntc:d}' )
    if verbose: even_print('max [t] chunk nt' , f'{max_ntc:d}' )
    if verbose: print(72*'-')
    
    # === init outfiles
    
    #if verbose: print(72*'*')
    if verbose: print(fn_h5_out)
    if verbose: print(len(fn_h5_out)*'-')
    
    ## initialize file: turbulent kinetic energy budget, AVERAGED
    #with rgd(fn_h5_out, 'w', force=force, driver='mpio', comm=self.comm) as f1:
    with rgd_meta(fn_h5_out, 'w', force=force, driver='mpio', comm=self.comm) as f1:
        
        f1.init_from_rgd(self.fname, t_info=False)
        
        ## set some top-level attributes
        #f1.attrs['duration_avg'] = duration_avg ## duration of mean
        f1.attrs['duration_avg'] = self.t[-1] - self.t[0]
        #f1_mean.attrs['duration_avg'] = self.duration
        f1.attrs['dt'] = self.t[1] - self.t[0]
        #f1.attrs['fclass'] = 'rgd'
        f1.attrs['fsubtype'] = 'mean'
        
        shape = (1,f1.nz,f1.ny,f1.nx)
        chunks = h5_chunk_sizer(nxi=shape, constraint=chunk_constraint, size_kb=chunk_kb, base=chunk_base, itemsize=8)
        
        data_gb = 8 * 1 * f1.nz*f1.ny*f1.nx / 1024**3
        
        for dss in ['production','dissipation','transport','diffusion','p_dilatation','p_diffusion']:
            
            if verbose:
                even_print('initializing data/%s'%(dss,),'%0.1f [GB]'%(data_gb,))
            
            dset = f1.create_dataset(
                        f'data/{dss}', 
                        shape=shape, 
                        dtype=np.float64,
                        chunks=chunks,
                        #data=np.full(shape,0.,np.float64),
                        )
            
            chunk_kb_ = np.prod(dset.chunks) * dset.dtype.itemsize / 1024. ## actual
            if verbose:
                even_print('chunk shape (t,z,y,x)','%s'%str(dset.chunks))
                even_print('chunk size','%i [KB]'%int(round(chunk_kb_)))
        
        # === replace dims/t array --> take last time of series
        if ('dims/t' in f1):
            del f1['dims/t']
        f1.create_dataset('dims/t', data=np.array([self.t[-1]],dtype=np.float64), )
        
        if hasattr(f1, 'duration_avg'):
            if verbose: even_print('duration_avg', '%0.2f'%f1.duration_avg)
    
    if verbose: print(72*'-')
    
    ## initialize file: turbulent kinetic energy budget, UNSTEADY
    if save_unsteady:
        
        if verbose: print(fn_h5_out_unsteady)
        if verbose: print(len(str(fn_h5_out_unsteady))*'-')
        
        #with rgd(fn_h5_out_unsteady, 'w', force=force, driver='mpio', comm=self.comm) as f1:
        with rgd_meta(fn_h5_out_unsteady, 'w', force=force, driver='mpio', comm=self.comm) as f1:
            
            f1.init_from_rgd(self.fname, t_info=True)
            
            ## set some top-level attributes
            #f1.attrs['duration_avg'] = duration_avg ## duration of mean
            f1.attrs['duration_avg'] = self.t[-1] - self.t[0]
            #f1_mean.attrs['duration_avg'] = self.duration
            f1.attrs['dt'] = self.t[1] - self.t[0]
            #f1.attrs['fclass'] = 'rgd'
            f1.attrs['fsubtype'] = 'unsteady'
            
            shape = (self.nt,f1.nz,f1.ny,f1.nx)
            #chunks = h5_chunk_sizer(nxi=shape, constraint=chunk_constraint, size_kb=chunk_kb, base=chunk_base, itemsize=4)
            chunks = self['data/u'].chunks ## use chunk pattern of this file
            
            data_gb = 4*self.nt*f1.nz*f1.ny*f1.nx / 1024**3
            
            for dss in ['production','dissipation','transport','diffusion']: #,'p_dilatation','p_diffusion']:
                
                if verbose:
                    even_print('initializing data/%s'%(dss,),'%0.1f [GB]'%(data_gb,))
                
                dset = f1.create_dataset(
                            f'data/{dss}', 
                            shape=shape, 
                            dtype=np.float32,
                            chunks=chunks,
                            #data=np.full(shape,0.,np.float32),
                            )
                
                chunk_kb_ = np.prod(dset.chunks) * dset.dtype.itemsize / 1024. ## actual
                if verbose:
                    even_print('chunk shape (t,z,y,x)','%s'%str(dset.chunks))
                    even_print('chunk size','%i [KB]'%int(round(chunk_kb_)))
        
        if verbose: print(72*'-')
    
    self.comm.Barrier()
    
    # ===
    
    ## OPEN the mean file... make sure to close later!
    #hf_mean = rgd(fn_h5_mean, 'r', driver='mpio', comm=self.comm)
    hf_mean = rgd_meta(fn_h5_mean, 'r', driver='mpio', comm=self.comm)
    
    ## verify contents of the mean file
    np.testing.assert_allclose( hf_mean.x       , self.x       , atol=1e-12, rtol=1e-12 )
    np.testing.assert_allclose( hf_mean.y       , self.y       , atol=1e-12, rtol=1e-12 )
    np.testing.assert_allclose( hf_mean.z       , self.z       , atol=1e-12, rtol=1e-12 )
    np.testing.assert_allclose( hf_mean.lchar   , self.lchar   , atol=1e-12, rtol=1e-12 )
    np.testing.assert_allclose( hf_mean.U_inf   , self.U_inf   , atol=1e-12, rtol=1e-12 )
    np.testing.assert_allclose( hf_mean.T_inf   , self.T_inf   , atol=1e-12, rtol=1e-12 )
    np.testing.assert_allclose( hf_mean.rho_inf , self.rho_inf , atol=1e-12, rtol=1e-12 )
    if (hf_mean.fsubtype!='mean'):
        raise ValueError
    
    # ===
    
    data = {} ## dict to be output to .dat file
    
    if ('data_dim' not in hf_mean):
        raise ValueError('group data_dim not present')
    
    ## put all data from 'data_dim' into the dictionary data which will be pickled at the end
    for dsn in hf_mean['data_dim'].keys():
        d_ = np.copy( hf_mean[f'data_dim/{dsn}'][()] )
        if (d_.ndim == 0):
            d_ = float(d_)
        data[dsn] = d_
    
    ## 1D
    #rho_avg = np.copy( hf_mean['data_dim/rho'][()] )
    #u_avg   = np.copy( hf_mean['data_dim/u'][()]   )
    #v_avg   = np.copy( hf_mean['data_dim/v'][()]   )
    #w_avg   = np.copy( hf_mean['data_dim/w'][()]   )
    #T_avg   = np.copy( hf_mean['data_dim/T'][()]   )
    #p_avg   = np.copy( hf_mean['data_dim/p'][()]   )
    
    ## 0D
    u_tau    = float( hf_mean['data_dim/u_tau'][()]    )
    nu_wall  = float( hf_mean['data_dim/nu_wall'][()]  )
    rho_wall = float( hf_mean['data_dim/rho_wall'][()] )
    #T_wall   = float( hf_mean['data_dim/T_wall'][()]   )
    d99      = float( hf_mean['data_dim/d99'][()]      )
    u_99     = float( hf_mean['data_dim/u_99'][()]     )
    Re_tau   = float( hf_mean['data_dim/Re_tau'][()]   )
    Re_theta = float( hf_mean['data_dim/Re_theta'][()] )
    #sc_u_in  = float( hf_mean['data_dim/sc_u_in'][()]  )
    sc_l_in  = float( hf_mean['data_dim/sc_l_in'][()]  )
    sc_t_in  = float( hf_mean['data_dim/sc_t_in'][()]  )
    #sc_u_out = float( hf_mean['data_dim/sc_u_out'][()] )
    #sc_l_out = float( hf_mean['data_dim/sc_l_out'][()] )
    #sc_t_out = float( hf_mean['data_dim/sc_t_out'][()] )
    
    ## 0D scalars
    lchar   = self.lchar   ; data['lchar']   = lchar
    U_inf   = self.U_inf   ; data['U_inf']   = U_inf
    rho_inf = self.rho_inf ; data['rho_inf'] = rho_inf
    T_inf   = self.T_inf   ; data['T_inf']   = T_inf
    
    #data['M_inf'] = self.M_inf
    data['Ma'] = self.Ma
    data['Pr'] = self.Pr
    
    ## read in 1D coordinate arrays & re-dimensionalize
    x = np.copy( self['dims/x'][()] * self.lchar )
    y = np.copy( self['dims/y'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    t = np.copy( self['dims/t'][()] * self.tchar )
    
    ## dimensional [s]
    dt = self.dt * self.tchar
    np.testing.assert_allclose(dt, t[1]-t[0], rtol=1e-14, atol=1e-14)
    
    t_meas = self.duration * self.tchar
    np.testing.assert_allclose(t_meas, t.max()-t.min(), rtol=1e-14, atol=1e-14)
    
    t_eddy = t_meas / ( d99 / u_tau )
    
    ## check if constant Δz (calculate Δz+ later)
    dz0 = np.diff(z)[0]
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-7)):
        raise NotImplementedError
    np.testing.assert_allclose(dz0, z[1]-z[0], rtol=1e-14, atol=1e-14)
    
    zrange = z.max() - z.min()
    np.testing.assert_allclose(zrange, z[-1]-z[0], rtol=1e-14, atol=1e-14)
    
    nx = self.nx ; data['nx'] = nx
    ny = self.ny ; data['ny'] = ny
    nz = self.nz ; data['nz'] = nz
    nt = self.nt ; data['nt'] = nt
    
    ## add data to dict that gets output
    data['x'] = x
    data['y'] = y
    data['z'] = z
    data['t'] = t
    data['t_meas'] = t_meas
    data['dt'] = dt
    data['dz0'] = dz0
    data['zrange'] = zrange
    
    if verbose: even_print('Δt/tchar','%0.8f'%(dt/self.tchar))
    if verbose: even_print('Δt','%0.3e [s]'%(dt,))
    if verbose: even_print('duration/tchar','%0.1f'%(self.duration,))
    if verbose: even_print('duration','%0.3e [s]'%(self.duration*self.tchar,))
    if verbose: print(72*'-')
    
    ## report
    if verbose:
        even_print('dt'     , f'{dt     :0.5e} [s]' )
        even_print('t_meas' , f'{t_meas :0.5e} [s]' )
        even_print('dz0'    , f'{dz0    :0.5e} [m]' )
        even_print('zrange' , f'{zrange :0.5e} [m]' )
        print(72*'-')
    
    ## report
    if verbose:
        even_print( 'Reτ'             , f'{Re_tau:0.1f}'           )
        even_print( 'Reθ'             , f'{Re_theta:0.1f}'         )
        even_print( 'δ99'             , f'{d99:0.5e} [m]'          )
        even_print( 'δν=(ν_wall/u_τ)' , f'{sc_l_in:0.5e} [m]'      )
        even_print( 'U_inf'           , f'{self.U_inf:0.3f} [m/s]' )
        even_print( 'uτ'              , f'{u_tau:0.3f} [m/s]'      )
        even_print( 'ν_wall'          , f'{nu_wall:0.5e} [m²/s]'   )
        even_print( 'ρ_wall'          , f'{rho_wall:0.6f} [kg/m³]' )
        even_print( 'Δz+'             , f'{dz0/sc_l_in:0.3f}'      )
        even_print( 'zrange/δ99'      , f'{zrange/d99:0.3f}'       )
        even_print( 'Δt+'             , f'{dt/sc_t_in:0.3f}'       )
        print(72*'-')
    
    ## report
    if verbose:
        even_print('t_meas/(δ99/u_τ) = t_eddy' , '%0.2f'%t_eddy)
        even_print('t_meas/(δ99/u99)'          , '%0.2f'%(t_meas/(d99/u_99)))
        even_print('t_meas/(20·δ99/u99)'       , '%0.2f'%(t_meas/(20*d99/u_99)))
        print(72*'-')
    
    # ===
    
    ## copy AVG [u,v,w] into memory... DIMENSIONAL
    u_re = np.copy( U_inf * hf_mean['data/u'][0,:,:,:].T    ).astype(np.float64)
    v_re = np.copy( U_inf * hf_mean['data/v'][0,:,:,:].T    ).astype(np.float64)
    w_re = np.copy( U_inf * hf_mean['data/w'][0,:,:,:].T    ).astype(np.float64)
    u_fv = np.copy( U_inf * hf_mean['data/u_fv'][0,:,:,:].T ).astype(np.float64)
    v_fv = np.copy( U_inf * hf_mean['data/v_fv'][0,:,:,:].T ).astype(np.float64)
    w_fv = np.copy( U_inf * hf_mean['data/w_fv'][0,:,:,:].T ).astype(np.float64)
    
    p_re = np.copy( rho_inf * U_inf**2 * hf_mean['data/p'][0,:,:,:].T ).astype(np.float64)
    
    ## get Reynolds avg strain tensor elements
    #dudx_re = gradient( u=u_re , x=x , axis=0 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    #dudy_re = gradient( u=u_re , x=y , axis=1 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    #dudz_re = gradient( u=u_re , x=z , axis=2 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    #dvdx_re = gradient( u=v_re , x=x , axis=0 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    #dvdy_re = gradient( u=v_re , x=y , axis=1 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    #dvdz_re = gradient( u=v_re , x=z , axis=2 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    #dwdx_re = gradient( u=w_re , x=x , axis=0 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    #dwdy_re = gradient( u=w_re , x=y , axis=1 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    #dwdz_re = gradient( u=w_re , x=z , axis=2 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    
    ## get Favre avg strain tensor elements
    dudx_fv = gradient( u=u_fv , x=x , axis=0 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    dudy_fv = gradient( u=u_fv , x=y , axis=1 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    dudz_fv = gradient( u=u_fv , x=z , axis=2 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    dvdx_fv = gradient( u=v_fv , x=x , axis=0 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    dvdy_fv = gradient( u=v_fv , x=y , axis=1 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    dvdz_fv = gradient( u=v_fv , x=z , axis=2 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    dwdx_fv = gradient( u=w_fv , x=x , axis=0 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    dwdy_fv = gradient( u=w_fv , x=y , axis=1 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    dwdz_fv = gradient( u=w_fv , x=z , axis=2 , acc=acc , edge_stencil=edge_stencil )[:,:,:,np.newaxis]
    
    ## accumulators for per-timestep sum, at end gets multiplied by (1/nt) to get average
    unsteady_production_sum   = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
    unsteady_dissipation_sum  = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
    unsteady_transport_sum    = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
    unsteady_diffusion_sum    = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
    unsteady_p_diffusion_sum  = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
    unsteady_p_dilatation_sum = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
    
    # ==============================================================
    # check memory
    # ==============================================================
    
    hostname = MPI.Get_processor_name()
    mem_free_gb = psutil.virtual_memory().free / 1024**3
    G = self.comm.gather([ self.rank , hostname , mem_free_gb ], root=0)
    G = self.comm.bcast(G, root=0)
    
    ## multiple ranks per node
    host_mem = {}
    for rank, host, mem in G:
        if (host not in host_mem) or (mem < host_mem[host]):
            host_mem[host] = mem
    total_free = sum(host_mem.values())
    
    if verbose:
        for key,value in host_mem.items():
            even_print(f'RAM free {key}', f'{int(np.floor(value)):d} [GB]')
        print(72*'-')
        even_print('RAM free (local,min)', f'{int(np.floor(min(host_mem.values()))):d} [GB]')
        even_print('RAM free (global)', f'{int(np.floor(total_free)):d} [GB]')
    
    shape_read = ( nx, ny, nz, max_ntc ) ## local
    if verbose: even_print('read shape (local,max)', f'[{nx:d},{ny:d},{nz:d},{max_ntc:d}]')
    data_gb = np.dtype(np.float64).itemsize * np.prod(shape_read) / 1024**3
    if verbose: even_print('read size (local)', f'{data_gb:0.2f} [GB]')
    if verbose: even_print('read size (global)', f'{int(np.ceil(data_gb*rt)):d} [GB]')
    
    fac = 65
    if verbose: even_print(f'read size (global) ×{fac:d}', f'{int(np.ceil(data_gb*rt*fac)):d} [GB]')
    ram_usage_est = data_gb*rt*fac/total_free
    if verbose: even_print('RAM usage estimate', f'{100*ram_usage_est:0.1f} [%]')
    
    self.comm.Barrier()
    if (ram_usage_est>0.90):
        print('RAM consumption might be too high. exiting.')
        self.comm.Abort(1)
    
    if verbose:
        print(72*'-')
    
    # ==============================================================
    # main loop
    # ==============================================================
    
    if verbose:
        progress_bar = tqdm(
            total=ct,
            ncols=100,
            desc='turb budget',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#FF6600',
            )
    
    ct_counter=0
    for ctl_ in ctl:
        ct_counter += 1
        ct1, ct2 = ctl_
        ntc = ct2 - ct1
        
        G = self.comm.gather([ self.rank , ntc ], root=0)
        G = self.comm.bcast(G, root=0)
        ntc_global = sum( [ g[1] for g in G ] )
        
        if (ct>1):
            if verbose:
                mesg = f'[t] sub chunk {ct_counter:d}/{ct:d}'
                tqdm.write( mesg )
                tqdm.write( '-'*len(mesg) )
                tqdm.write( even_print('nt in chunk', f'{ntc_global:d}', s=True) )
        
        # === read unsteady data
        
        ss1 = ['u','v','w','T','rho','p']
        ss2 = ['uI','vI','wI', 'uII','vII','wII', 'pI'] ## 'TI','rhoI'
        
        ## data buffer
        formats = [ np.float64 for s in ss1+ss2 ]
        shape   = (nx,ny,nz,ntc)
        dd      = np.zeros(shape=shape, dtype={'names':ss1+ss2, 'formats':formats}, order='C')
        
        ## read DIMLESS u,v,w,T,rho,p
        for ss in ss1:
            dset = self['data/%s'%ss]
            self.comm.Barrier()
            t_start = timeit.default_timer()
            with dset.collective:
                dd[ss] = dset[ct1:ct2,:,:,:].T
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            data_gb = dset.dtype.itemsize * nx*ny*nz * ntc_global / 1024**3
            if verbose:
                tqdm.write( even_print('read: %s'%ss, '%0.3f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True) )
        
        ## re-dimensionalize unsteady data
        dd['u'][:,:,:,:]   *= U_inf
        dd['v'][:,:,:,:]   *= U_inf
        dd['w'][:,:,:,:]   *= U_inf
        dd['T'][:,:,:,:]   *= T_inf
        dd['rho'][:,:,:,:] *= rho_inf
        dd['p'][:,:,:,:]   *= ( rho_inf * U_inf**2 )
        
        ## calculate DIMENSIONAL primes
        dd['uI'][:,:,:,:]  = dd['u'] - u_re[:,:,:,np.newaxis]
        dd['vI'][:,:,:,:]  = dd['v'] - v_re[:,:,:,np.newaxis]
        dd['wI'][:,:,:,:]  = dd['w'] - w_re[:,:,:,np.newaxis]
        dd['uII'][:,:,:,:] = dd['u'] - u_fv[:,:,:,np.newaxis]
        dd['vII'][:,:,:,:] = dd['v'] - v_fv[:,:,:,np.newaxis]
        dd['wII'][:,:,:,:] = dd['w'] - w_fv[:,:,:,np.newaxis]
        dd['pI'][:,:,:,:]  = dd['p'] - p_re[:,:,:,np.newaxis]
        
        if verbose: tqdm.write(72*'-')
        
        data_gb = dd.nbytes / 1024**3
        if verbose:
            tqdm.write(even_print('unsteady data (rank)', f'{data_gb:0.1f} [GB]', s=True))
        
        # === make VIEWS of the numpy structured array for convenience
        
        #u   = dd['u'][:,:,:,:]
        #v   = dd['v'][:,:,:,:]
        #w   = dd['w'][:,:,:,:]
        T   = dd['T'][:,:,:,:]
        #p   = dd['p'][:,:,:,:]
        rho = dd['rho'][:,:,:,:]
        
        uI  = dd['uI'][:,:,:,:]
        vI  = dd['vI'][:,:,:,:]
        wI  = dd['wI'][:,:,:,:]
        uII = dd['uII'][:,:,:,:]
        vII = dd['vII'][:,:,:,:]
        wII = dd['wII'][:,:,:,:]
        pI  = dd['pI'][:,:,:,:]
        
        mu = np.copy( self.C_Suth * T**(3/2) / (T + self.S_Suth) )
        #nu = np.copy( mu / rho )
        
        # dd = None ; del dd
        
        self.comm.Barrier()
        mem_free_gb = psutil.virtual_memory().free/1024**3
        if verbose:
            tqdm.write(even_print('mem free', f'{mem_free_gb:0.1f} [GB]', s=True))
        
        # === get gradients
        
        t_start = timeit.default_timer() ## rank 0 only
        
        #dudx   = gradient(u, x, acc=acc, edge_stencil=edge_stencil, axis=0)
        #dudy   = gradient(u, y, acc=acc, edge_stencil=edge_stencil, axis=1)
        #dudz   = gradient(u, z, acc=acc, edge_stencil=edge_stencil, axis=2)
        #dvdx   = gradient(v, x, acc=acc, edge_stencil=edge_stencil, axis=0)
        #dvdy   = gradient(v, y, acc=acc, edge_stencil=edge_stencil, axis=1)
        #dvdz   = gradient(v, z, acc=acc, edge_stencil=edge_stencil, axis=2)
        #dwdx   = gradient(w, x, acc=acc, edge_stencil=edge_stencil, axis=0)
        #dwdy   = gradient(w, y, acc=acc, edge_stencil=edge_stencil, axis=1)
        #dwdz   = gradient(w, z, acc=acc, edge_stencil=edge_stencil, axis=2)
        
        duIdx  = gradient(uI, x, acc=acc, edge_stencil=edge_stencil, axis=0)
        duIdy  = gradient(uI, y, acc=acc, edge_stencil=edge_stencil, axis=1)
        duIdz  = gradient(uI, z, acc=acc, edge_stencil=edge_stencil, axis=2)
        dvIdx  = gradient(vI, x, acc=acc, edge_stencil=edge_stencil, axis=0)
        dvIdy  = gradient(vI, y, acc=acc, edge_stencil=edge_stencil, axis=1)
        dvIdz  = gradient(vI, z, acc=acc, edge_stencil=edge_stencil, axis=2)
        dwIdx  = gradient(wI, x, acc=acc, edge_stencil=edge_stencil, axis=0)
        dwIdy  = gradient(wI, y, acc=acc, edge_stencil=edge_stencil, axis=1)
        dwIdz  = gradient(wI, z, acc=acc, edge_stencil=edge_stencil, axis=2)
        
        duIIdx = gradient(uII, x, acc=acc, edge_stencil=edge_stencil, axis=0)
        duIIdy = gradient(uII, y, acc=acc, edge_stencil=edge_stencil, axis=1)
        duIIdz = gradient(uII, z, acc=acc, edge_stencil=edge_stencil, axis=2)
        dvIIdx = gradient(vII, x, acc=acc, edge_stencil=edge_stencil, axis=0)
        dvIIdy = gradient(vII, y, acc=acc, edge_stencil=edge_stencil, axis=1)
        dvIIdz = gradient(vII, z, acc=acc, edge_stencil=edge_stencil, axis=2)
        dwIIdx = gradient(wII, x, acc=acc, edge_stencil=edge_stencil, axis=0)
        dwIIdy = gradient(wII, y, acc=acc, edge_stencil=edge_stencil, axis=1)
        dwIIdz = gradient(wII, z, acc=acc, edge_stencil=edge_stencil, axis=2)
        
        t_delta = timeit.default_timer() - t_start ## rank 0 only
        
        if verbose:
            tqdm.write(even_print('get gradients', format_time_string(t_delta), s=True))
        
        self.comm.Barrier()
        mem_free_gb  = psutil.virtual_memory().free/1024**3
        if verbose:
            tqdm.write(even_print('mem free', f'{mem_free_gb:0.1f} [GB]', s=True))
        
        # === stack velocities into vectors and mean-removed strains into tensors
        
        t_start = timeit.default_timer()
        
        uI_i  = np.stack(( uI  , vI  , wI  ), axis=-1)
        uII_i = np.stack(( uII , vII , wII ), axis=-1)
        
        duIdx_ij = np.stack((np.stack((duIdx, duIdy, duIdz), axis=4),
                             np.stack((dvIdx, dvIdy, dvIdz), axis=4),
                             np.stack((dwIdx, dwIdy, dwIdz), axis=4)), axis=5)
        
        duIIdx_ij = np.stack((np.stack((duIIdx, duIIdy, duIIdz), axis=4),
                              np.stack((dvIIdx, dvIIdy, dvIIdz), axis=4),
                              np.stack((dwIIdx, dwIIdy, dwIIdz), axis=4)), axis=5)
        
        t_delta = timeit.default_timer() - t_start
        if verbose:
            tqdm.write(even_print('tensor stacking',format_time_string(t_delta), s=True))
        
        self.comm.Barrier()
        mem_free_gb = psutil.virtual_memory().free/1024**3
        if verbose:
            tqdm.write(even_print('mem free', f'{mem_free_gb:0.1f} [GB]', s=True))
        
        # === production : P
        if True:
            
            if verbose: tqdm.write(72*'-')
            t_start = timeit.default_timer()
            
            r_uII_uII = rho*uII*uII
            r_uII_vII = rho*uII*vII
            r_uII_wII = rho*uII*wII
            
            r_vII_uII = rho*vII*uII
            r_vII_vII = rho*vII*vII
            r_vII_wII = rho*vII*wII
            
            r_wII_uII = rho*wII*uII
            r_wII_vII = rho*wII*vII
            r_wII_wII = rho*wII*wII
            
            ## unsteady_production_ = - ( r_uII_uII * dudx_fv + r_uII_vII * dudy_fv + r_uII_wII * dudz_fv \
            ##                          + r_uII_vII * dvdx_fv + r_vII_vII * dvdy_fv + r_vII_wII * dvdz_fv \
            ##                          + r_uII_wII * dwdx_fv + r_vII_wII * dwdy_fv + r_wII_wII * dwdz_fv )
            
            r_uIIuII_ij = np.stack((np.stack((r_uII_uII, r_uII_vII, r_uII_wII), axis=4),
                                    np.stack((r_vII_uII, r_vII_vII, r_vII_wII), axis=4),
                                    np.stack((r_wII_uII, r_wII_vII, r_wII_wII), axis=4)), axis=5)
            
            dudx_fv_ij = np.stack((np.stack((dudx_fv, dudy_fv, dudz_fv), axis=4),
                                   np.stack((dvdx_fv, dvdy_fv, dvdz_fv), axis=4),
                                   np.stack((dwdx_fv, dwdy_fv, dwdz_fv), axis=4)), axis=5)
            
            unsteady_production = -1*np.einsum('xyztij,xyztij->xyzt', r_uIIuII_ij, dudx_fv_ij)
            
            ## np.testing.assert_allclose(unsteady_production, unsteady_production_, atol=20000)
            ## print('check passed : np.einsum()')
            
            t_delta = timeit.default_timer() - t_start
            if verbose:
                tqdm.write(even_print('calc production', format_time_string(t_delta), s=True))
            
            self.comm.Barrier()
            
            if save_unsteady: ## write 4D unsteady production
                
                ## technically this is an estimate
                data_gb = self.n_ranks * np.prod(unsteady_production.shape) * unsteady_production.dtype.itemsize / 1024**3
                
                #with rgd(fn_h5_out_unsteady, 'a', driver='mpio', comm=self.comm) as f1:
                with rgd_meta(fn_h5_out_unsteady, 'a', driver='mpio', comm=self.comm) as f1:
                    dset = f1['data/production']
                    self.comm.Barrier()
                    t_start = timeit.default_timer()
                    with dset.collective:
                        dset[ct1:ct2,:,:,:] = unsteady_production.T
                    self.comm.Barrier()
                    t_delta = timeit.default_timer() - t_start
                    if verbose:
                        tqdm.write(even_print('write: production', '%0.2f [GB]  %0.2f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
            
            ## combine sums across ranks for this [t] chunk
            unsteady_production_sum_i   = np.sum(unsteady_production, axis=3, keepdims=True, dtype=np.float64)
            unsteady_production_sum_buf = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
            self.comm.Reduce(
                        [unsteady_production_sum_i, MPI.DOUBLE],
                        [unsteady_production_sum_buf, MPI.DOUBLE],
                        op=MPI.SUM,
                        root=0,
                        )
            
            ## add aggregated sum (across ranks) to the total accumulator
            unsteady_production_sum += unsteady_production_sum_buf
            
            # === release mem
            r_uII_uII           = None; del r_uII_uII
            r_uII_vII           = None; del r_uII_vII
            r_uII_wII           = None; del r_uII_wII
            r_vII_uII           = None; del r_vII_uII
            r_vII_vII           = None; del r_vII_vII
            r_vII_wII           = None; del r_vII_wII
            r_wII_uII           = None; del r_wII_uII
            r_wII_vII           = None; del r_wII_vII
            r_wII_wII           = None; del r_wII_wII
            r_uIIuII_ij         = None; del r_uIIuII_ij
            dudx_fv_ij          = None; del dudx_fv_ij
            unsteady_production = None; del unsteady_production
            
            self.comm.Barrier()
            mem_free_gb = psutil.virtual_memory().free/1024**3
            if verbose:
                tqdm.write(even_print('mem free', f'{mem_free_gb:0.1f} [GB]', s=True))
        
        # === dissipation : ε
        if True:
            
            if verbose: tqdm.write(72*'-')
            t_start = timeit.default_timer()
            
            duIIdx_ij_duIdx_ij = np.einsum('xyztij,xyztij->xyzt', duIIdx_ij, duIdx_ij)
            duIIdx_ij_duIdx_ji = np.einsum('xyztij,xyztji->xyzt', duIIdx_ij, duIdx_ij)
            
            ## # === from pp_turbulent_budget.F90
            ## unsteady_dissipation_ = mu * ( (duIdx + duIdx) * duIIdx  +  (duIdy + dvIdx) * duIIdy  +  (duIdz + dwIdx) * duIIdz \
            ##                              + (dvIdx + duIdy) * dvIIdx  +  (dvIdy + dvIdy) * dvIIdy  +  (dvIdz + dwIdy) * dvIIdz \
            ##                              + (dwIdx + duIdz) * dwIIdx  +  (dwIdy + dvIdz) * dwIIdy  +  (dwIdz + dwIdz) * dwIIdz )
            
            unsteady_dissipation = mu*(duIIdx_ij_duIdx_ij + duIIdx_ij_duIdx_ji)
            
            ## np.testing.assert_allclose(unsteady_dissipation, unsteady_dissipation_, rtol=1e-4)
            ## print('check passed : np.einsum() : dissipation')
            
            t_delta = timeit.default_timer() - t_start
            if verbose:
                tqdm.write(even_print('calc dissipation', format_time_string(t_delta),s=True))
            
            self.comm.Barrier()
            
            if save_unsteady: ## write 4D unsteady dissipation
                
                ## technically this is an estimate
                data_gb = self.n_ranks * np.prod(unsteady_dissipation.shape) * unsteady_dissipation.dtype.itemsize / 1024**3
                
                #with rgd(fn_h5_out_unsteady, 'a', driver='mpio', comm=self.comm) as f1:
                with rgd_meta(fn_h5_out_unsteady, 'a', driver='mpio', comm=self.comm) as f1:
                    dset = f1['data/dissipation']
                    self.comm.Barrier()
                    t_start = timeit.default_timer()
                    with dset.collective:
                        dset[ct1:ct2,:,:,:] = unsteady_dissipation.T
                    self.comm.Barrier()
                    t_delta = timeit.default_timer() - t_start
                    if verbose:
                        tqdm.write(even_print('write: dissipation', '%0.2f [GB]  %0.2f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
            
            ## combine sums across ranks for this [t] chunk
            unsteady_dissipation_sum_i   = np.sum(unsteady_dissipation, axis=3, keepdims=True, dtype=np.float64)
            unsteady_dissipation_sum_buf = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
            self.comm.Reduce(
                        [unsteady_dissipation_sum_i, MPI.DOUBLE],
                        [unsteady_dissipation_sum_buf, MPI.DOUBLE],
                        op=MPI.SUM,
                        root=0,
                        )
            
            ## add aggregated sum (across ranks) to the total accumulator
            unsteady_dissipation_sum += unsteady_dissipation_sum_buf
            
            ## release mem
            duIIdx_ij_duIdx_ij   = None; del duIIdx_ij_duIdx_ij
            duIIdx_ij_duIdx_ji   = None; del duIIdx_ij_duIdx_ji
            unsteady_dissipation = None; del unsteady_dissipation
            
            self.comm.Barrier()
            mem_free_gb = psutil.virtual_memory().free/1024**3
            if verbose:
                tqdm.write(even_print('mem free', f'{mem_free_gb:0.1f} [GB]', s=True))
        
        # === transport : T
        if True:
            
            if verbose: tqdm.write(72*'-')
            t_start = timeit.default_timer()
            
            ## triple correlation
            tc = np.einsum('xyzt,xyzti,xyzti,xyztj->xyztj', rho, uII_i, uII_i, uII_i)
            
            #tc_ddx = np.gradient(tc, x, axis=0, edge_order=2)
            #tc_ddy = np.gradient(tc, y, axis=1, edge_order=2)
            #tc_ddz = np.gradient(tc, z, axis=2, edge_order=2)
            
            tc_ddx = gradient(tc, x, axis=0, acc=acc, edge_stencil=edge_stencil)
            tc_ddy = gradient(tc, y, axis=1, acc=acc, edge_stencil=edge_stencil)
            tc_ddz = gradient(tc, z, axis=2, acc=acc, edge_stencil=edge_stencil)
            
            unsteady_transport = -0.5*(tc_ddx[:,:,:,:,0] + tc_ddy[:,:,:,:,1] + tc_ddz[:,:,:,:,2])
            
            t_delta = timeit.default_timer() - t_start
            if verbose:
                tqdm.write(even_print('calc transport', format_time_string(t_delta),s=True))
            
            self.comm.Barrier()
            
            if save_unsteady: ## write 4D unsteady transport
                
                ## technically this is an estimate
                data_gb = self.n_ranks * np.prod(unsteady_transport.shape) * unsteady_transport.dtype.itemsize / 1024**3
                
                #with rgd(fn_h5_out_unsteady, 'a', driver='mpio', comm=self.comm) as f1:
                with rgd_meta(fn_h5_out_unsteady, 'a', driver='mpio', comm=self.comm) as f1:
                    dset = f1['data/transport']
                    self.comm.Barrier()
                    t_start = timeit.default_timer()
                    with dset.collective:
                        dset[ct1:ct2,:,:,:] = unsteady_transport.T
                    self.comm.Barrier()
                    t_delta = timeit.default_timer() - t_start
                    if verbose:
                        tqdm.write(even_print('write: transport', '%0.2f [GB]  %0.2f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
            
            ## combine sums across ranks for this [t] chunk
            unsteady_transport_sum_i   = np.sum(unsteady_transport, axis=3, keepdims=True, dtype=np.float64)
            unsteady_transport_sum_buf = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
            self.comm.Reduce(
                        [unsteady_transport_sum_i, MPI.DOUBLE],
                        [unsteady_transport_sum_buf, MPI.DOUBLE],
                        op=MPI.SUM,
                        root=0,
                        )
            
            ## add aggregated sum (across ranks) to the total accumulator
            unsteady_transport_sum += unsteady_transport_sum_buf
            
            ## release mem
            tc     = None; del tc
            tc_ddx = None; del tc_ddx
            tc_ddy = None; del tc_ddy
            tc_ddz = None; del tc_ddz
            unsteady_transport = None; del unsteady_transport
            
            self.comm.Barrier()
            mem_free_gb = psutil.virtual_memory().free/1024**3
            if verbose:
                tqdm.write(even_print('mem free', f'{mem_free_gb:0.1f} [GB]', s=True))
        
        # === viscous diffusion : D
        if True:
            
            if verbose: tqdm.write(72*'-')
            t_start = timeit.default_timer()
            
            omega_ij = duIdx_ij + np.transpose(duIdx_ij, axes=(0,1,2,3,5,4))
            
            if False:
                
                omega_ij_2 = np.stack((np.stack(((duIdx+duIdx), (dvIdx+duIdy), (dwIdx+duIdz)), axis=4),
                                       np.stack(((duIdy+dvIdx), (dvIdy+dvIdy), (dwIdy+dvIdz)), axis=4),
                                       np.stack(((duIdz+dwIdx), (dvIdz+dwIdy), (dwIdz+dwIdz)), axis=4)), axis=5)
                
                np.testing.assert_allclose(omega_ij, omega_ij_2, rtol=1e-8)
                
                if verbose:
                    print('check passed : omega_ij')
            
            A = np.einsum('xyzt,xyzti,xyztij->xyztj', mu, uI_i, omega_ij)
            
            A_ddx = gradient(A[:,:,:,:,0], x, axis=0, acc=acc, edge_stencil=edge_stencil)
            A_ddy = gradient(A[:,:,:,:,1], y, axis=1, acc=acc, edge_stencil=edge_stencil)
            A_ddz = gradient(A[:,:,:,:,2], z, axis=2, acc=acc, edge_stencil=edge_stencil)
            
            unsteady_diffusion = A_ddx + A_ddy + A_ddz
            
            t_delta = timeit.default_timer() - t_start
            if verbose:
                tqdm.write(even_print('calc diffusion', format_time_string(t_delta),s=True))
            
            self.comm.Barrier()
            
            if save_unsteady: ## write 4D unsteady diffusion
                
                ## technically this is an estimate
                data_gb = self.n_ranks * np.prod(unsteady_diffusion.shape) * unsteady_diffusion.dtype.itemsize / 1024**3
                
                #with rgd(fn_h5_out_unsteady, 'a', driver='mpio', comm=self.comm) as f1:
                with rgd_meta(fn_h5_out_unsteady, 'a', driver='mpio', comm=self.comm) as f1:
                    dset = f1['data/diffusion']
                    self.comm.Barrier()
                    t_start = timeit.default_timer()
                    with dset.collective:
                        dset[ct1:ct2,:,:,:] = unsteady_diffusion.T
                    self.comm.Barrier()
                    t_delta = timeit.default_timer() - t_start
                    if verbose:
                        tqdm.write(even_print('write: diffusion', '%0.2f [GB]  %0.2f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True))
            
            ## combine sums across ranks for this [t] chunk
            unsteady_diffusion_sum_i   = np.sum(unsteady_diffusion, axis=3, keepdims=True, dtype=np.float64)
            unsteady_diffusion_sum_buf = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
            self.comm.Reduce(
                        [unsteady_diffusion_sum_i, MPI.DOUBLE],
                        [unsteady_diffusion_sum_buf, MPI.DOUBLE],
                        op=MPI.SUM,
                        root=0,
                        )
            
            ## add aggregated sum (across ranks) to the total accumulator
            unsteady_diffusion_sum += unsteady_diffusion_sum_buf
            
            ## release mem
            omega_ij = None; del omega_ij
            A        = None; del A
            A_ddx    = None; del A_ddx
            A_ddy    = None; del A_ddy
            A_ddz    = None; del A_ddz
            unsteady_diffusion = None; del unsteady_diffusion
            
            self.comm.Barrier()
            mem_free_gb = psutil.virtual_memory().free/1024**3
            if verbose:
                tqdm.write(even_print('mem free', f'{mem_free_gb:0.1f} [GB]', s=True))
        
        # === pressure diffusion
        if True:
            
            if verbose: tqdm.write(72*'-')
            t_start = timeit.default_timer()
            
            A = np.einsum('xyzti,xyzt->xyzti', uII_i, pI)
            
            #A_ddx = np.gradient(A[:,:,:,:,0], x, axis=0, edge_order=2)
            #A_ddy = np.gradient(A[:,:,:,:,1], y, axis=1, edge_order=2)
            #A_ddz = np.gradient(A[:,:,:,:,2], z, axis=2, edge_order=2)
            
            A_ddx = gradient(A[:,:,:,:,0], x, axis=0, acc=acc, edge_stencil=edge_stencil)
            A_ddy = gradient(A[:,:,:,:,1], y, axis=1, acc=acc, edge_stencil=edge_stencil)
            A_ddz = gradient(A[:,:,:,:,2], z, axis=2, acc=acc, edge_stencil=edge_stencil)
            
            unsteady_p_diffusion = A_ddx + A_ddy + A_ddz
            
            t_delta = timeit.default_timer() - t_start
            if verbose:
                tqdm.write(even_print('calc p_diffusion', format_time_string(t_delta),s=True))
            
            self.comm.Barrier()
            
            ## combine sums across ranks for this [t] chunk
            unsteady_p_diffusion_sum_i   = np.sum(unsteady_p_diffusion, axis=3, keepdims=True, dtype=np.float64)
            unsteady_p_diffusion_sum_buf = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
            self.comm.Reduce(
                            [unsteady_p_diffusion_sum_i, MPI.DOUBLE],
                            [unsteady_p_diffusion_sum_buf, MPI.DOUBLE],
                            op=MPI.SUM,
                            root=0,
                            )
            
            ## add aggregated sum (across ranks) to the total accumulator
            unsteady_p_diffusion_sum += unsteady_p_diffusion_sum_buf
            
            # === release mem
            A        = None; del A
            A_ddx    = None; del A_ddx
            A_ddy    = None; del A_ddy
            A_ddz    = None; del A_ddz
            unsteady_p_diffusion = None; del unsteady_p_diffusion
            
            self.comm.Barrier()
            mem_free_gb = psutil.virtual_memory().free/1024**3
            if verbose:
                tqdm.write(even_print('mem free', f'{mem_free_gb:0.1f} [GB]', s=True))
        
        # === pressure dilatation
        if True:
            
            if verbose: tqdm.write(72*'-')
            t_start = timeit.default_timer()
            
            # A = np.einsum('xyzt,xyzti->xyzti', pI, uII_i)
            # A_ddx = np.gradient(A[:,:,:,:,0], x, axis=0, edge_order=2)
            # A_ddy = np.gradient(A[:,:,:,:,1], y, axis=1, edge_order=2)
            # A_ddz = np.gradient(A[:,:,:,:,2], z, axis=2, edge_order=2)
            # unsteady_p_dilatation = A_ddx + A_ddy + A_ddz
            
            unsteady_p_dilatation = pI * ( duIIdx + dvIIdy + dwIIdz )
            
            t_delta = timeit.default_timer() - t_start
            if verbose:
                tqdm.write(even_print('calc p_dilatation', format_time_string(t_delta),s=True))
            
            self.comm.Barrier()
            
            ## combine sums across ranks for this [t] chunk
            unsteady_p_dilatation_sum_i   = np.sum(unsteady_p_dilatation, axis=3, keepdims=True, dtype=np.float64)
            unsteady_p_dilatation_sum_buf = np.zeros((nx,ny,nz,1), dtype=np.float64, order='C')
            self.comm.Reduce(
                            [unsteady_p_dilatation_sum_i, MPI.DOUBLE],
                            [unsteady_p_dilatation_sum_buf, MPI.DOUBLE],
                            op=MPI.SUM,
                            root=0,
                            )
            
            ## add aggregated sum (across ranks) to the total accumulator
            unsteady_p_dilatation_sum += unsteady_p_dilatation_sum_buf
            
            ## release mem
            unsteady_p_dilatation = None
            del unsteady_p_dilatation
            
            self.comm.Barrier()
            mem_free_gb = psutil.virtual_memory().free/1024**3
            if verbose:
                tqdm.write(even_print('mem free', f'{mem_free_gb:0.1f} [GB]', s=True))
        
        if verbose:
            progress_bar.update()
            tqdm.write(72*'-')
        
        #break ## debug
    
    if verbose: progress_bar.close()
    if verbose: print(72*'-')
    
    ## CLOSE the files that were opened
    hf_mean.close()
    
    # ==============================================================
    # multiply accumulators by (1/n) to get [t] avg
    # ==============================================================
    
    if (self.rank==0):
        production   = np.copy( ((1/nt) * unsteady_production_sum   ) )
        dissipation  = np.copy( ((1/nt) * unsteady_dissipation_sum  ) )
        transport    = np.copy( ((1/nt) * unsteady_transport_sum    ) )
        diffusion    = np.copy( ((1/nt) * unsteady_diffusion_sum    ) )
        p_diffusion  = np.copy( ((1/nt) * unsteady_p_diffusion_sum  ) )
        p_dilatation = np.copy( ((1/nt) * unsteady_p_dilatation_sum ) )
    self.comm.Barrier()
    
    # === write to HDF5
    
    if self.rank==0:
        #with rgd(fn_h5_out, 'a') as f1:
        with rgd_meta(fn_h5_out, 'a') as f1:
            f1['data/production'][:,:,:,:]   = production.T
            f1['data/dissipation'][:,:,:,:]  = dissipation.T
            f1['data/transport'][:,:,:,:]    = transport.T
            f1['data/diffusion'][:,:,:,:]    = diffusion.T
            f1['data/p_diffusion'][:,:,:,:]  = p_diffusion.T
            f1['data/p_dilatation'][:,:,:,:] = p_dilatation.T
    self.comm.Barrier()
    
    ## report file contents
    if (self.rank==0):
        with h5py.File(fn_h5_out,'r') as hfr:
            h5_print_contents(hfr)
    self.comm.Barrier()
    
    if verbose: print(72*'-')
    if verbose: print('total time : rgd.calc_turb_budget() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    
    ## make .xdmf
    if self.rank==0:
        #with rgd(fn_h5_out, 'r') as f1:
        with rgd_meta(fn_h5_out, 'r') as f1:
            f1.make_xdmf()
    self.comm.Barrier()
    
    ## make .xdmf
    if save_unsteady:
        if self.rank==0:
            #with rgd(fn_h5_out_unsteady, 'r') as f1:
            with rgd_meta(fn_h5_out_unsteady, 'r') as f1:
                f1.make_xdmf()
        self.comm.Barrier()
    
    return



import os
import sys
import timeit
from pathlib import Path, PurePosixPath

import h5py
import numpy as np
import scipy as sp
from scipy.signal import csd
from tqdm import tqdm

from .h5 import h5_print_contents
from .utils import even_print, format_time_string

# ======================================================================

def _calc_turb_cospectrum_wall(self, **kwargs):
    '''
    Calculate FFT cospectrum in [z,t] at every [x]
    - designed for analyzing unsteady, pre-computed wall quantities ([y] plane)
    '''
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'spd.calc_turb_cospectrum_wall()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    if not self.usingmpi:
        raise NotImplementedError('function is not implemented for non-MPI usage')
    
    h5py_is_mpi_build = h5py.h5.get_config().mpi
    if not h5py_is_mpi_build:
        if verbose: print('h5py was not compiled for parallel usage! exiting.')
        sys.exit(1)
    
    ri = kwargs.get('ri',1)
    rj = kwargs.get('rj',1)
    rt = kwargs.get('rt',1)
    
    fn_h5_out       = kwargs.get('fn_h5_out',None)       ## filename for output HDF5 (.h5) file
    overlap_fac_nom = kwargs.get('overlap_fac_nom',0.50) ## nominal windows overlap factor
    n_win           = kwargs.get('n_win',8)              ## number of segment windows for [t] PSD calc
    
    ## check data distribution (only distribute data in [i]/[x])
    if (rj!=1):
        raise AssertionError('rj!=1')
    if (rt!=1):
        raise AssertionError('rt!=1')
    
    if not isinstance(ri,int) or (ri<1):
        raise ValueError('ri should be a positive non-zero int')
    
    if (ri*rj*rt != self.n_ranks):
        raise AssertionError('ri*rj*rt != self.n_ranks')
    if (ri>self.ni):
        raise AssertionError('ri>self.ni')
    if (self.ni%ri!=0):
        raise AssertionError('ni currently needs to be divisible by the n ranks')
    
    ## distribute data over [i]/[x]
    ril_ = np.array_split(np.arange(self.ni,dtype=np.int64),min(ri,self.ni))
    ril = [[b[0],b[-1]+1] for b in ril_ ]
    ri1,ri2 = ril[self.rank]
    nir = ri2 - ri1
    
    ## output filename : HDF5 (.h5)
    if (fn_h5_out is None): ## automatically determine name
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        #fname_root = re.findall(r'io\S+_mpi_[0-9]+', fname_root)[0]
        fn_h5_out_base = fname_root+'_fft.h5'
        fn_h5_out = str(PurePosixPath(fname_path, fn_h5_out_base))
    if (Path(fn_h5_out).suffix != '.h5'):
        raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' must end in .h5")
    if os.path.isfile(fn_h5_out):
        if (fn_h5_out == self.fname):
            raise ValueError(f"fn_h5_out='{str(fn_h5_out)}' cannot be same as input filename.")
    
    if verbose: even_print( 'fn_h5'     , self.fname )
    if verbose: even_print( 'fn_h5_out' , fn_h5_out  )
    if verbose: print(72*'-')
    self.comm.Barrier()
    
    # ===
    
    ## the data dictionary to be written to .h5 later
    data = {}
    
    ## infile
    fsize = os.path.getsize(self.fname)/1024**3
    if verbose: even_print(os.path.basename(self.fname),'%0.1f [GB]'%fsize)
    if verbose: even_print('ni',f'{self.ni:d}')
    if verbose: even_print('nj',f'{self.nj:d}')
    if verbose: even_print('nt',f'{self.nt:d}')
    if verbose: even_print('n_pts',f'{self.n_pts/1e6:0.1f} [M]')
    if verbose: even_print('n_ranks',f'{self.n_ranks:d}')
    if verbose: print(72*'-')
    
    ## 0D freestream scalars
    #lchar   = self.lchar   ; data['lchar']   = self.lchar
    U_inf   = self.U_inf   ; data['U_inf']   = self.U_inf
    rho_inf = self.rho_inf ; data['rho_inf'] = self.rho_inf
    T_inf   = self.T_inf   ; data['T_inf']   = self.T_inf
    #mu_inf  = self.mu_inf  ; data['mu_inf']  = self.mu_inf
    data['p_inf'] = self.p_inf
    #data['M_inf'] = self.M_inf
    data['Ma']    = self.Ma
    data['Pr']    = self.Pr
    
    ## read in 1D coordinate arrays & re-dimensionalize
    x = np.copy( self['dims/x'][()] * self.lchar )
    z = np.copy( self['dims/z'][()] * self.lchar )
    t = np.copy( self['dims/t'][()] * self.tchar )
    data['x'] = x
    data['z'] = z
    data['t'] = t
    
    nx = self.ni
    ni = self.ni
    nz = self.nj
    nj = self.nj
    data['ni'] = ni
    data['nx'] = nx
    data['nj'] = nj
    data['nz'] = nz
    
    nt = self.nt
    data['nt'] = nt
    
    ## assert constant Δz
    dz0 = np.diff(z)[0]
    if not np.all(np.isclose(np.diff(z), dz0, rtol=1e-6)):
        raise NotImplementedError('Δz not constant')
    
    ## get Δt, dimensional [s]
    if hasattr(self,'dt'):
        dt = self.dt * self.tchar
        np.testing.assert_allclose(dt, t[1]-t[0], rtol=1e-12, atol=1e-12)
    else:
        dt = t[1] - t[0] ## already dimensional
    
    if hasattr(self,'duration'):
        t_meas = self.duration * self.tchar
        np.testing.assert_allclose(t_meas, t.max()-t.min(), rtol=1e-12, atol=1e-12)
    else:
        t_meas = t[-1] - t[0] ## already dimensional
        self.duration = t_meas / self.tchar ## non-dimensionalize for attribute
    
    zrange = z.max() - z.min()
    
    data['t'] = t
    data['dt'] = dt
    data['dz'] = dz0
    data['zrange'] = zrange
    
    if verbose: even_print( 'Δt/tchar'       , f'{dt/self.tchar:0.8f}' )
    if verbose: even_print( 'Δt'             , f'{dt:0.3e} [s]'        )
    if verbose: even_print( 'duration/tchar' , f'{self.duration:0.1f}' )
    if verbose: even_print( 'duration'       , f'{self.duration*self.tchar:0.3e} [s]' )
    if verbose: print(72*'-')
    
    ## report
    if verbose:
        even_print('Δt'     , f'{dt    :0.5e} [s]' )
        even_print('t_meas' , f'{t_meas:0.5e} [s]' )
        even_print('Δz'     , f'{dz0   :0.5e} [m]' )
        even_print('zrange' , f'{zrange:0.5e} [m]' )
        print(72*'-')
    
    ## establish [t] windowing & get frequency
    nperseg     = nt // n_win
    noverlap    = int(round(nperseg*overlap_fac_nom))
    overlap_fac = noverlap / nperseg
    fs          = 1./dt ## dimensional [1/s]
    
    ## get [freq] vector
    freq,_ = csd(
            np.zeros((nt,),dtype=np.float64),
            np.zeros((nt,),dtype=np.float64),
            fs=fs,
            nperseg=nperseg,
            noverlap=noverlap,
            window='hann',
            detrend='constant',
            scaling='density',
            return_onesided=True,
            )
    
    nf = freq.shape[0]
    df = np.diff(freq)[0]
    
    data['nperseg']  = nperseg
    data['noverlap'] = noverlap
    data['freq']     = freq
    data['df']       = df
    data['nf']       = nf
    
    if verbose:
        even_print('overlap_fac (nominal)' , f'{overlap_fac_nom:0.5f}'  )
        even_print('n_win'                 , f'{n_win:d}'               )
        even_print('nperseg'               , f'{nperseg:d}'             )
        even_print('noverlap'              , f'{noverlap:d}'            )
        even_print('overlap_fac'           , f'{overlap_fac:0.5f}'      )
        print(72*'-')
    
    if verbose:
        even_print('freq min',f'{freq.min():0.1f} [Hz]')
        even_print('freq max',f'{freq.max():0.1f} [Hz]')
        even_print('df',f'{df:0.1f} [Hz]')
        even_print('nf',f'{nf:d}')
        print(72*'-')
    
    ## spatial [z] wavenumber (kz) vector
    kz_full = sp.fft.fftfreq(n=nz, d=dz0) * ( 2 * np.pi )
    kzp     = np.where(kz_full>0) ## indices of positive values
    kz      = np.copy(kz_full[kzp])
    dkz     = kz[1] - kz[0]
    nkz     = kz.shape[0]
    
    ## wavenumber vector should be size nz//2-1
    if (nkz!=nz//2-1):
        raise ValueError
    
    data['kz']  = kz
    data['dkz'] = dkz
    data['nkz'] = nkz
    
    if verbose:
        even_print('kz min',f'{kz.min():0.1f} [1/m]')
        even_print('kz max',f'{kz.max():0.1f} [1/m]')
        even_print('dkz',f'{dkz:0.1f} [1/m]')
        even_print('nkz',f'{nkz:d}')
        print(72*'-')
    
    ## wavelength λz = (2·π)/kz
    lz = np.copy( 2 * np.pi / kz )
    data['lz'] = lz
    
    # ==============================================================
    # prepare buffers, etc.
    # ==============================================================
    
    do_density_weighting = False ## deactivated for now... would need implementation
    
    ## cospectrum pairs
    ## [ str:var1, str:var2 ]
    fft_combis = [
    
    [ 'u_tau' , 'u_tau' ], ## [ uτ , uτ ] --> FFT[ uτ′ , uτ′ ]
    [ 'v_tau' , 'v_tau' ], ## [ vτ , vτ ] --> FFT[ vτ′ , vτ′ ]
    [ 'w_tau' , 'w_tau' ], ## [ wτ , wτ ] --> FFT[ wτ′ , wτ′ ]
    
    #[ 'u_tau' , 'v_tau' ], ## [ uτ , vτ ] --> FFT[ uτ′ , vτ′ ]
    #[ 'u_tau' , 'w_tau' ], ## [ uτ , wτ ] --> FFT[ uτ′ , wτ′ ]
    #[ 'v_tau' , 'w_tau' ], ## [ vτ , wτ ] --> FFT[ vτ′ , wτ′ ]
    
    [ 'tau_uy' , 'tau_uy' ], ## [ τuy , τuy ] --> FFT[ τuy′ , τuy′ ]
    [ 'tau_vy' , 'tau_vy' ], ## [ τvy , τvy ] --> FFT[ τvy′ , τvy′ ]
    [ 'tau_wy' , 'tau_wy' ], ## [ τwy , τwy ] --> FFT[ τwy′ , τwy′ ]
    
    #[ 'tau_uy' , 'tau_vy' ], ## [ τuy , τvy ] --> FFT[ τuy′ , τvy′ ]
    #[ 'tau_uy' , 'tau_wy' ], ## [ τuy , τwy ] --> FFT[ τuy′ , τwy′ ]
    #[ 'tau_vy' , 'tau_wy' ], ## [ τvy , τwy ] --> FFT[ τvy′ , τwy′ ]
    
    [ 'p'   , 'p'   ], ## [p,p] --> FFT[ p′ , p′ ]
    [ 'T'   , 'T'   ], ## [T,T] --> FFT[ T′ , T′ ]
    [ 'rho' , 'rho' ], ## [ρ,ρ] --> FFT[ ρ′ , ρ′ ]
    
    [ 'u_tau'  , 'p' ], ## [uτ,p] --> FFT[ uτ′  , p′ ]
    [ 'tau_uy' , 'p' ], ## [τuy,p] --> FFT[ τuy′ , p′ ]
    
    ]
    
    ## generate FFT cospectrum scalar names
    scalars = []
    for fft_combi in fft_combis:
        #s1,s2,do_density_weighting = fft_combi
        s1,s2 = fft_combi
        if do_density_weighting:
            raise NotImplementedError
        else:
            scalars.append(f"{s1.replace('_','')}I_{s2.replace('_','')}I")
    
    ## generate AVG scalar names
    scalars_Re_avg = []
    #scalars_Fv_avg = []
    for fft_combi in fft_combis:
        #s1,s2,do_density_weighting = fft_combi
        s1,s2 = fft_combi
        if do_density_weighting and ('rho' not in scalars_Re_avg):
            scalars_Re_avg.append('rho')
        if do_density_weighting:
            #if (s1 not in scalars_Fv_avg):
            #    scalars_Fv_avg.append(s1)
            #if (s2 not in scalars_Fv_avg):
            #    scalars_Fv_avg.append(s2)
            raise NotImplementedError
        else:
            if (s1 not in scalars_Re_avg):
                scalars_Re_avg.append(s1)
            if (s2 not in scalars_Re_avg):
                scalars_Re_avg.append(s2)
    
    ## numpy formatted arrays: buffers for PSD & other data (rank-local)
    Ekz        = np.zeros(shape=(nir,nkz ) , dtype={'names':scalars         , 'formats':[ np.dtype(np.complex128) for s in scalars        ]})
    Ef         = np.zeros(shape=(nir,nf  ) , dtype={'names':scalars         , 'formats':[ np.dtype(np.complex128) for s in scalars        ]})
    covariance = np.zeros(shape=(nir,    ) , dtype={'names':scalars         , 'formats':[ np.dtype(np.float64)    for s in scalars        ]})
    avg_Re     = np.zeros(shape=(nir,    ) , dtype={'names':scalars_Re_avg  , 'formats':[ np.dtype(np.float64)    for s in scalars_Re_avg ]})
    #avg_Fv     = np.zeros(shape=(nir,    ) , dtype={'names':scalars_Fv_avg  , 'formats':[ np.dtype(np.float64)    for s in scalars_Fv_avg ]})
    
    if verbose:
        even_print('n turb spectrum scalar combinations' , '%i'%(len(fft_combis),))
        print(72*'-')
    
    ## window for [z] -- rectangular because [z] is assumed periodic already
    window_z       = np.ones(nz,dtype=np.float64)
    mean_sq_win_z  = np.mean(window_z**2)
    if verbose:
        even_print('mean(window_z**2)', '%0.5f'%(mean_sq_win_z,))
    
    # ==============================================================
    # main loop
    # ==============================================================
    
    if verbose:
        progress_bar = tqdm(
            total=len(fft_combis)*nir,
            ncols=100,
            desc='fft',
            leave=True,
            file=sys.stdout,
            mininterval=0.1,
            smoothing=0.,
            #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
            bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
            ascii="░█",
            colour='#FF6600',
            )
    
    for cci,cc in enumerate(fft_combis): ## fft pairs
        
        if verbose: tqdm.write(72*'-')
        
        scalar_L, scalar_R = cc
        
        msg = f'FFT[{scalar_L}′,{scalar_R}′]'
        if verbose:
            tqdm.write(even_print('computing',msg,s=True,))
        
        dset_L = self[f'data/{scalar_L}']
        dset_R = self[f'data/{scalar_R}']
        data_gb_1x = self.n_ranks * 1 * self.nj * self.nt * dset_L.dtype.itemsize / 1024**3
        
        scalar = scalars[cci]
        
        ## assert scalar name
        scalar_ = f"{scalar_L.replace('_','')}I_{scalar_R.replace('_','')}I"
        if (scalar != scalar_):
            raise ValueError
        
        ## assert scalar name
        if do_density_weighting:
            #if (f'r{scalar_L}II_r{scalar_R}II' != scalar ):
            #    raise ValueError
            raise NotImplementedError
        else:
            if (f"{scalar_L.replace('_','')}I_{scalar_R.replace('_','')}I" != scalar ):
                raise ValueError
        
        ## [x] loop (rank-local)
        for ii in range(ri1,ri2):
            
            iii = ii - ri1
            
            data_L = np.zeros( (nz,nt) , dtype=np.float64 )
            data_R = np.zeros( (nz,nt) , dtype=np.float64 )
            
            self.comm.Barrier()
            t_start = timeit.default_timer()
            
            ## read data L
            n_scalars_read = 1 ## initialize
            scalar_str = scalar_L ## initialize
            with dset_L.collective:
                data_L[:,:] = np.copy( dset_L[ii,:,:] ).astype(np.float64)
            self.comm.Barrier()
            
            ## read data R (if != data L)
            if (scalar_L==scalar_R):
                data_R[:,:] = np.copy( data_L )
            else:
                n_scalars_read += 1
                scalar_str += f',{scalar_R}'
                with dset_R.collective:
                    data_R[:,:] = np.copy( dset_R[ii,:,:] ).astype(np.float64)
                self.comm.Barrier()
            
            self.comm.Barrier()
            t_delta = timeit.default_timer() - t_start
            data_gb = data_gb_1x * n_scalars_read
            
            if verbose:
                tqdm.write(even_print(f'read: {scalar_str}', f'{data_gb:0.3f} [GB]  {t_delta:0.3f} [s]  {data_gb/t_delta:0.3f} [GB/s]', s=True))
            
            ## data_L and data_R shape should be (nz,nt)
            if ( data_L.shape != (nz,nt) ) or ( data_R.shape != (nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            # === redimensionalize
            
            if scalar_L in ['tau_uy','tau_vy','tau_wy',]:
                data_L *= rho_inf * U_inf**2
            elif scalar_L in ['u_tau','v_tau','w_tau',]:
                data_L *= U_inf
            elif scalar_L in ['p',]:
                data_L *= rho_inf * U_inf**2
            elif scalar_L in ['T',]:
                data_L *= T_inf
            elif scalar_L in ['rho',]:
                data_L *= rho_inf
            else:
                raise ValueError
            
            if scalar_R in ['tau_uy','tau_vy','tau_wy',]:
                data_R *= rho_inf * U_inf**2
            elif scalar_R in ['u_tau','v_tau','w_tau',]:
                data_R *= U_inf
            elif scalar_R in ['p',]:
                data_R *= rho_inf * U_inf**2
            elif scalar_R in ['T',]:
                data_R *= T_inf
            elif scalar_R in ['rho',]:
                data_R *= rho_inf
            else:
                raise ValueError
            
            ## data_L and data_R shape should be (nz,nt)
            if ( data_L.shape != (nz,nt) ) or ( data_R.shape != (nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            # === compute mean-removed data
            
            ## avg(□) or avg(ρ·□)/avg(ρ) in [t]
            if do_density_weighting:
                #rho_avg     = np.mean(        rho , axis=-1, dtype=np.float64, keepdims=True)
                #data_L_avg  = np.mean( rho*data_L , axis=-1, dtype=np.float64, keepdims=True)
                #data_L_avg /= rho_avg
                #data_R_avg  = np.mean( rho*data_R , axis=-1, dtype=np.float64, keepdims=True)
                #data_R_avg /= rho_avg
                raise NotImplementedError
            else:
                data_L_avg = np.mean( data_L , axis=-1, dtype=np.float64, keepdims=True) ## (nz,1)
                data_R_avg = np.mean( data_R , axis=-1, dtype=np.float64, keepdims=True) ## (nz,1)
            
            ## data_L_avg and data_R_avg shape should be (nz,1)
            if ( data_L_avg.shape != (nz,1) ) or ( data_R_avg.shape != (nz,1) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## Reynolds prime □′ or Favre prime □″ --> shape (nz,nt)
            data_L -= data_L_avg
            data_R -= data_R_avg
            
            ## data_L and data_R shape should be (nz,nt)
            if ( data_L.shape != (nz,nt) ) or ( data_R.shape != (nz,nt) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## assert stationarity / definition averaging
            ## avg(□′)==0 or avg(ρ·□″)==0
            if do_density_weighting:
                #a_ = np.mean(rho*data_L, axis=-1, dtype=np.float64, keepdims=True)
                #b_ = np.mean(rho*data_R, axis=-1, dtype=np.float64, keepdims=True)
                raise NotImplementedError
            else:
                a_ = np.mean(data_L, axis=-1, dtype=np.float64, keepdims=True) ## average in [t] --> (nz,1)
                b_ = np.mean(data_R, axis=-1, dtype=np.float64, keepdims=True)
            if not np.allclose( a_, np.zeros_like(a_), atol=1e-6 ) or not np.allclose( b_, np.zeros_like(b_), atol=1e-6 ):
                print(f'rank {self.rank:d}: avg(□′)!=0 or avg(ρ·□″)!=0')
                self.comm.Abort(1)
            
            ## covariance: <□′·□′> OR <ρ□″·ρ□″> --> note that this is NOT the typical Favre <ρ·□″□″>
            if do_density_weighting:
                #covariance_ = np.mean( rho*data_L * rho*data_R , axis=-1 , dtype=np.float64, keepdims=True)
                raise NotImplementedError
            else:
                covariance_ = np.mean( data_L*data_R , axis=-1 , dtype=np.float64, keepdims=True) ## average in [t] --> (nz,1)
            
            if ( covariance_.shape != (nz,1) ):
                print(f'rank {self.rank:d}: shape violation')
                self.comm.Abort(1)
            
            ## write this chunk/scalar's covariance to covariance buffer
            ## avg over [z,t] --> np.float64()
            covariance[scalar][iii] = np.mean( covariance_ , axis=(0,1) , dtype=np.float64)
            
            ## write (rank-local) 1D [x] average
            if do_density_weighting:
                #avg_Fv[scalar_L][iii] = np.mean( data_L_avg , axis=(0,1) , dtype=np.float64) )
                #avg_Fv[scalar_R][iii] = np.mean( data_R_avg , axis=(0,1) , dtype=np.float64) )
                #avg_Re['rho'][iii]    = np.mean( rho_avg    , axis=(0,1) , dtype=np.float64) )
                raise ValueError
            else:
                avg_Re[scalar_L][iii] = np.mean( data_L_avg , axis=(0,1) , dtype=np.float64)
                avg_Re[scalar_R][iii] = np.mean( data_R_avg , axis=(0,1) , dtype=np.float64)
            
            # ===============================================================================
            # At this point you have 4D [x,y,z,t] □′ data
            # ===============================================================================
            
            ## do [z] FFT for every [t]
            Ekz_buf = np.zeros((nt,nkz), dtype=np.complex128)
            for ti in range(nt):
                uL = np.copy( data_L[:,ti] )
                uR = np.copy( data_R[:,ti] )
                
                ## One-sided amplitude spectra
                A1 = sp.fft.fft( uL * window_z )[kzp] / nz
                A2 = sp.fft.fft( uR * window_z )[kzp] / nz
                
                Ekz_buf[ti,:] = 2. * A1 * np.conj(A2) / ( dkz * mean_sq_win_z )
            
            Ekz[scalar][iii,:] = np.mean(Ekz_buf, axis=0) ## mean across [t] --> (nkz,)
            
            ## do [t] FFT for every [z]
            Ef_buf = np.zeros((nz,nf), dtype=np.complex128)
            #for zi in range(self.nj):
            for zi in range(nz):
                uL = np.copy( data_L[zi,:] )
                uR = np.copy( data_R[zi,:] )
                
                ## One-sided complex cross-spectral density in [f] (using Welch's method)
                _,P = csd(
                        uL,uR,
                        fs=fs,
                        nperseg=nperseg,
                        noverlap=noverlap,
                        window='hann',
                        detrend='constant',
                        scaling='density',
                        return_onesided=True,
                        )
                
                Ef_buf[zi,:] = P
            
            Ef[scalar][iii,:] = np.mean(Ef_buf, axis=0) ## mean across [z] --> (nf,)
            
            self.comm.Barrier() ## [x] loop ('ii' within this rank's range)
            if verbose: progress_bar.update()
            
            #break ## DEBUG
    
    self.comm.Barrier()
    if verbose:
        progress_bar.close()
        print(72*'-')
    
    # ==============================================================
    # write HDF5 (.h5) file
    # ==============================================================
    
    ## overwrite outfile!
    ## open on rank 0 and write attributes, dimensions, etc.
    if (self.rank==0):
        with h5py.File(fn_h5_out, 'w') as hfw:
            
            ## write floats,ints as top-level attributes
            for key,val in data.items():
                if isinstance(data[key], (int,np.int32,np.int64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], (float,np.float32,np.float64)):
                    hfw.attrs[key] = val
                elif isinstance(data[key], np.ndarray):
                    pass
                else:
                    print(f'key {key} is type {str(type(data[key]))}')
                    self.comm.Abort(1)
            
            ## write numpy arrays
            hfw.create_dataset( 'dims/x'    , data=x    ) ## [m]
            hfw.create_dataset( 'dims/z'    , data=z    ) ## [m]
            hfw.create_dataset( 'dims/t'    , data=t    ) ## [s]
            hfw.create_dataset( 'dims/freq' , data=freq ) ## [1/s] | [Hz]
            hfw.create_dataset( 'dims/kz'   , data=kz   ) ## [1/m]
            hfw.create_dataset( 'dims/lz'   , data=lz   ) ## [m]
            
            ## initialize datasets : covariance,Ekz,Ef
            for scalar in scalars:
                hfw.create_dataset( f'covariance/{scalar}' , shape=(nx,)    , dtype=np.float64    , chunks=None    , data=np.full((nx,),0.,np.float64)       )
                hfw.create_dataset( f'Ekz/{scalar}'        , shape=(nx,nkz) , dtype=np.complex128 , chunks=(1,nkz) , data=np.full((nx,nkz),0.,np.complex128) )
                hfw.create_dataset( f'Ef/{scalar}'         , shape=(nx,nf)  , dtype=np.complex128 , chunks=(1,nf)  , data=np.full((nx,nf),0.,np.complex128)  )
            
            ## initialize datasets : 1D [x] mean
            for scalar in avg_Re.dtype.names:
                hfw.create_dataset( f'avg/Re/{scalar}', shape=(nx,), dtype=np.float64, chunks=None, data=np.full((nx,),0.,np.float64) )
    
    self.comm.Barrier()
    
    ## re-open in parallel for data writes
    with h5py.File(fn_h5_out, 'a', driver='mpio', comm=self.comm) as hfw:
        
        ## collectively write covariance,Ekz,Ef
        for scalar in scalars:
            dset = hfw[f'covariance/{scalar}']
            with dset.collective:
                dset[ri1:ri2] = covariance[scalar][:]
            dset = hfw[f'Ekz/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = Ekz[scalar][:,:]
            dset = hfw[f'Ef/{scalar}']
            with dset.collective:
                dset[ri1:ri2,:] = Ef[scalar][:,:]
        
        ## collectively write 1D [y] avgs
        for scalar in avg_Re.dtype.names:
            dset = hfw[f'avg/Re/{scalar}']
            with dset.collective:
                dset[ri1:ri2] = avg_Re[scalar][:]
    
    ## report file contents
    self.comm.Barrier()
    if (self.rank==0):
        even_print( os.path.basename(fn_h5_out) , f'{(os.path.getsize(fn_h5_out)/1024**2):0.1f} [MB]' )
        print(72*'-')
        with h5py.File(fn_h5_out,'r') as hfr:
            h5_print_contents(hfr)
    self.comm.Barrier()
    
    if verbose: print(72*'-')
    if verbose: print('total time : spd.calc_turb_cospectrum_wall() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import io
import struct

import numpy as np

from .utils import even_print

# ======================================================================

class eas3:
    '''
    Interface class for EAS3 files (legacy binary NS3D output format)
    - only binary read is supported, no writing
    '''
    
    def __init__(self, fname, **kwargs):
        '''
        initialize class instance
        '''
        self.fname   = fname
        self.verbose = kwargs.get('verbose',True)
        
        if isinstance(fname, str):
            self.f = open(fname,'rb')
        elif isinstance(fname, io.BytesIO):
            self.f = fname
        else:
            raise TypeError('fname should be type str or io.BytesIO')
        
        self.udef = self.get_header()
    
    def __enter__(self):
        '''
        for use with python 'with' statement
        '''
        #print('opening from enter() --> used with statement')
        return self
    
    def __exit__(self, exception_type, exception_value, exception_traceback):
        '''
        for use with python 'with' statement
        '''
        self.f.close()
        
        if exception_type is not None:
            print('\nsafely closed EAS3 due to exception')
            print(72*'-')
            print('exception type : '+str(exception_type))
        if exception_value is not None:
            print('exception_value : '+str(exception_value))
        if exception_traceback is not None:
            print('exception_traceback : '+str(exception_traceback))
        if exception_type is not None:
            print(72*'-')
    
    def close(self):
        '''
        close() passthrough to handle from binary open()
        '''
        self.f.close()
    
    def get_header(self, **kwargs):
        
        ATTRLEN     = kwargs.get('ATTRLEN',10)
        UDEFLEN     = kwargs.get('UDEFLEN',20)
        ChangeGmode = kwargs.get('ChangeGmode',1)
        
        ## Definitions
        #EAS2=1; EAS3=2
        IEEES=1; IEEED=2; IEEEQ=3
        #EAS3_NO_ATTR=1
        EAS3_ALL_ATTR=2
        EAS3_NO_G=1
        EAS3_X0DX_G=2
        #EAS3_UDEF_G=3
        EAS3_ALL_G=4
        #EAS3_FULL_G=5
        #EAS3_NO_UDEF=1
        EAS3_ALL_UDEF=2
        #EAS3_INT_UDEF=3
        
        self.IEEES = IEEES
        self.IEEED = IEEED
        self.IEEEQ = IEEEQ
        
        ## Identifier: 20 byte character
        #identifier = self.f.read(20).strip()
        
        ## File type: 8 byte integer
        #file_type = struct.unpack('!q',self.f.read(8))[0]
        
        ## Accuracy: 8 byte integer
        accuracy = struct.unpack('!q',self.f.read(8))[0]
        self.accuracy = accuracy
        
        if (self.accuracy == self.IEEES):
            self.dtype = np.float32
        elif (self.accuracy == self.IEEED):
            self.dtype = np.float64
        elif (self.accuracy == self.IEEEQ):
            self.dtype = np.float128
        else:
            raise ValueError('precision not identifiable')
        
        ## Array sizes: each 8 byte integer
        nzs   = struct.unpack('!q',self.f.read(8))[0]
        npar  = struct.unpack('!q',self.f.read(8))[0]
        ndim1 = struct.unpack('!q',self.f.read(8))[0]
        ndim2 = struct.unpack('!q',self.f.read(8))[0]
        ndim3 = struct.unpack('!q',self.f.read(8))[0]
        
        self.ndim1 = ndim1
        self.ndim2 = ndim2
        self.ndim3 = ndim3
        self.npar  = npar
        self.nzs   = nzs
        
        ## Attribute mode: 8 byte integer
        attribute_mode = struct.unpack('!q',self.f.read(8))[0]
        
        ## Geometry mode: each 8 byte integer
        gmode_time  = struct.unpack('!q',self.f.read(8))[0]
        gmode_param = struct.unpack('!q',self.f.read(8))[0]
        gmode_dim1  = struct.unpack('!q',self.f.read(8))[0]  
        gmode_dim2  = struct.unpack('!q',self.f.read(8))[0]
        gmode_dim3  = struct.unpack('!q',self.f.read(8))[0]
        
        ## Array sizes for geometry data: each 8 byte integer
        size_time  = struct.unpack('!q',self.f.read(8))[0]
        size_param = struct.unpack('!q',self.f.read(8))[0]
        size_dim1  = struct.unpack('!q',self.f.read(8))[0]
        size_dim2  = struct.unpack('!q',self.f.read(8))[0]
        size_dim3  = struct.unpack('!q',self.f.read(8))[0]
        
        ## Specification of user defined data: 8 byte integer
        udef = struct.unpack('!q',self.f.read(8))[0]
        
        ## Array sizes for used defined data: each 8 byte integer
        udef_char_size = struct.unpack('!q',self.f.read(8))[0]
        udef_int_size  = struct.unpack('!q',self.f.read(8))[0]
        udef_real_size = struct.unpack('!q',self.f.read(8))[0]
        
        ## Time step array: nzs x 8 byte
        time_step = np.zeros(nzs,int)
        for it in range(nzs):
            time_step[it] = struct.unpack('!q',self.f.read(8))[0]
        
        if attribute_mode==EAS3_ALL_ATTR:
            ## Time step attributes
            attr_time = [ self.f.read(ATTRLEN).decode('UTF-8').strip() ]
            for it in range(1,nzs):
                attr_time.append( self.f.read(ATTRLEN).decode('UTF-8').strip() )
            ## Parameter attributes
            attr_param = [ self.f.read(ATTRLEN).decode('UTF-8').strip() ]
            for it in range(1,npar):
                attr_param.append( self.f.read(ATTRLEN).decode('UTF-8').strip() )
            
            # Spatial attributes
            attr_dim1 = self.f.read(ATTRLEN).decode('UTF-8').strip()
            attr_dim2 = self.f.read(ATTRLEN).decode('UTF-8').strip()
            attr_dim3 = self.f.read(ATTRLEN).decode('UTF-8').strip()
        
        ## If geometry mode > EAS3_NO_G for time
        if gmode_time == EAS3_X0DX_G:
            time_data = np.zeros(2)
            for it in range(2):
                time_data[it] = struct.unpack('!d',self.f.read(8))[0]
        elif gmode_time == EAS3_ALL_G:
            time_data = np.zeros(size_time)
            for it in range(size_time):
                time_data[it] = struct.unpack('!d',self.f.read(8))[0]
        else:
            time_data = np.zeros(1)
        
        ## If geometry mode > EAS3_NO_G for parameters
        if gmode_param > EAS3_NO_G:
            param = np.zeros(size_param)
            for it in range(size_param):
                param[it] = struct.unpack('!d',self.f.read(8))[0]
        
        ## If geometry mode > EAS3_NO_G for dimensions 1 to 3
        dim1_data = np.zeros(size_dim1)
        if gmode_dim1 > EAS3_NO_G:
            for it in range(size_dim1):
                dim1_data[it] = struct.unpack('!d',self.f.read(8))[0]
            if abs(dim1_data[0]) < 1e-18: dim1_data[0] = 0.   
        dim2_data = np.zeros(size_dim2)
        if gmode_dim2 > EAS3_NO_G:
            for it in range(size_dim2):
                dim2_data[it] = struct.unpack('!d',self.f.read(8))[0]
            if abs(dim2_data[0]) < 1e-18: dim2_data[0] = 0.
        dim3_data = np.zeros(size_dim3)
        if gmode_dim3 > EAS3_NO_G:
            for it in range(size_dim3):
                dim3_data[it] = struct.unpack('!d',self.f.read(8))[0]
        else:
            dim3_data = None
        
        ## If user-defined data is chosen 
        if udef==EAS3_ALL_UDEF:
            udef_char = []
            for it in range(udef_char_size):
                udef_char.append(self.f.read(UDEFLEN).decode('UTF-8').strip())
            udef_int = np.zeros(udef_int_size,int)
            for it in range(udef_int_size):
                udef_int[it] = struct.unpack('!q',self.f.read(8))[0]
            udef_real = np.zeros(udef_real_size)
            for it in range(udef_real_size):
                udef_real[it] = struct.unpack('!d',self.f.read(8))[0]
        
        ## Option: convert gmode=EAS3_X0DX_G to gmode=EAS3_ALL_G
        if ChangeGmode==1:
            if gmode_dim1==EAS3_X0DX_G:
                dim1_data = np.linspace(dim1_data[0],dim1_data[0]+dim1_data[1]*(ndim1-1), ndim1)
                gmode_dim1=EAS3_ALL_G
            if gmode_dim2==EAS3_X0DX_G:
                dim2_data = np.linspace(dim2_data[0],dim2_data[0]+dim2_data[1]*(ndim2-1), ndim2)
                gmode_dim2=EAS3_ALL_G
            if gmode_dim3==EAS3_X0DX_G:
                dim3_data = np.linspace(dim3_data[0],dim3_data[0]+dim3_data[1]*(ndim3-1), ndim3)
                gmode_dim3=EAS3_ALL_G
            if gmode_time==EAS3_X0DX_G:
                time_data = np.linspace(time_data[0],time_data[0]+time_data[1]*(nzs  -1), nzs  )
                gmode_time=EAS3_ALL_G
        
        # ===
        
        self.attr_param = attr_param
        self.scalars    = attr_param
        self.t          = time_data
        self.nt         = self.t.size
        
        if (attr_dim1=='x'):
            self.x = dim1_data
        elif (attr_dim1=='y'):
            self.y = dim1_data
        elif (attr_dim1=='z'):
            self.z = dim1_data
        else:
            raise ValueError('attr_dim1 = %s not identifiable as any x,y,z'%attr_dim1)
        
        if (attr_dim2=='x'):
            self.x = dim2_data
        elif (attr_dim2=='y'):
            self.y = dim2_data
        elif (attr_dim2=='z'):
            self.z = dim2_data
        else:
            raise ValueError('attr_dim2 = %s not identifiable as any x,y,z'%attr_dim2)
        
        if (attr_dim3=='x'):
            self.x = dim3_data
        elif (attr_dim3=='y'):
            self.y = dim3_data
        elif (attr_dim3=='z'):
            self.z = dim3_data
        else:
            raise ValueError('attr_dim3 = %s not identifiable as any x,y,z'%attr_dim3)
        
        # === transpose order to [x,y,z]
        
        if all([(attr_dim1=='x'),(attr_dim2=='y'),(attr_dim3=='z')]):
            self.axes_transpose_xyz = (0,1,2)
        elif all([(attr_dim1=='y'),(attr_dim2=='x'),(attr_dim3=='z')]):
            self.axes_transpose_xyz = (1,0,2)
        elif all([(attr_dim1=='z'),(attr_dim2=='y'),(attr_dim3=='x')]):
            self.axes_transpose_xyz = (2,1,0)
        elif all([(attr_dim1=='x'),(attr_dim2=='z'),(attr_dim3=='y')]):
            self.axes_transpose_xyz = (0,2,1)
        elif all([(attr_dim1=='y'),(attr_dim2=='z'),(attr_dim3=='x')]):
            self.axes_transpose_xyz = (2,0,1)
        elif all([(attr_dim1=='z'),(attr_dim2=='x'),(attr_dim3=='y')]):
            self.axes_transpose_xyz = (1,2,0)
        else:
            raise ValueError('could not determine axes transpose')
        
        # ===
        
        self.nx  = self.x.shape[0]
        self.ny  = self.y.shape[0]
        self.nz  = self.z.shape[0]
        self.ngp = self.nx*self.ny*self.nz
        
        if self.verbose: print(72*'-')
        if self.verbose: even_print('nx', '%i'%self.nx )
        if self.verbose: even_print('ny', '%i'%self.ny )
        if self.verbose: even_print('nz', '%i'%self.nz )
        if self.verbose: even_print('ngp', '%i'%self.ngp )
        if self.verbose: print(72*'-')
        
        if self.verbose: even_print('x_min', '%0.2f'%self.x.min())
        if self.verbose: even_print('x_max', '%0.2f'%self.x.max())
        if self.verbose: even_print('dx begin : end', '%0.3E : %0.3E'%( (self.x[1]-self.x[0]), (self.x[-1]-self.x[-2]) ))
        if self.verbose: even_print('y_min', '%0.2f'%self.y.min())
        if self.verbose: even_print('y_max', '%0.2f'%self.y.max())
        if self.verbose: even_print('dy begin : end', '%0.3E : %0.3E'%( (self.y[1]-self.y[0]), (self.y[-1]-self.y[-2]) ))
        if self.verbose: even_print('z_min', '%0.2f'%self.z.min())
        if self.verbose: even_print('z_max', '%0.2f'%self.z.max())        
        if self.verbose: even_print('dz begin : end', '%0.3E : %0.3E'%( (self.z[1]-self.z[0]), (self.z[-1]-self.z[-2]) ))
        if self.verbose: print(72*'-'+'\n')
        
        # ===
        
        ## make a python dict from udef vectors
        udef_dict = {}
        for i in range(len(udef_char)):
            if (udef_char[i]!=''):
                if (udef_int[i]!=0):
                    udef_dict[udef_char[i]] = int(udef_int[i])
                elif (udef_real[i]!=0.):
                    udef_dict[udef_char[i]] = float(udef_real[i])
                else:
                    udef_dict[udef_char[i]] = 0.
        
        if self.verbose:
            print('udef from EAS3\n' + 72*'-')
            for key in udef_dict:
                if isinstance(udef_dict[key],float):
                    even_print(key, '%0.8f'%udef_dict[key])
                elif isinstance(udef_dict[key],int):
                    even_print(key, '%i'%udef_dict[key])
                else:
                    #print(type(udef_dict[key]))
                    raise TypeError('udef dict item not float or int')
            print(72*'-'+'\n')
        
        self.Ma    = udef_dict['Ma']
        self.Re    = udef_dict['Re']
        self.Pr    = udef_dict['Pr']
        self.kappa = udef_dict['kappa']
        self.R     = udef_dict['R']
        
        ## get T_inf
        if ('T_unend' in udef_dict):
            self.T_inf = udef_dict['T_unend']
        elif ('T_inf' in udef_dict):
            self.T_inf = udef_dict['T_inf']
        elif ('Tinf' in udef_dict):
            self.T_inf = udef_dict['Tinf']
        else:
            raise ValueError('T_inf not found in udef')
        
        ## get p_inf
        if ('p_unend' in udef_dict):
            self.p_inf = udef_dict['p_unend']
        elif ('p_inf' in udef_dict):
            self.p_inf = udef_dict['p_inf']
        elif ('pinf' in udef_dict):
            self.p_inf = udef_dict['pinf']
        else:
            raise ValueError('p_inf not found in udef')
        
        ## !!! import what is called 'C_Suth' in NS3D as 'S_Suth' !!!
        self.S_Suth = udef_dict['C_Suth'] ## [K] --> Sutherland temperature, usually 110.4 [K] for air
        
        self.mu_Suth_ref = udef_dict['mu_Suth_ref'] ## usually 1.716e-05 [kg/(m·s)] --> μ of air at T_Suth_ref = 273.15 [K]
        self.T_Suth_ref  = udef_dict['T_Suth_ref'] ## usually 273.15 [K]
        
        self.C_Suth = self.mu_Suth_ref/(self.T_Suth_ref**(3/2))*(self.T_Suth_ref + self.S_Suth) ## ~1.458e-6 [kg/(m·s·√K)]
        
        ## derived values
        self.rho_inf = self.p_inf/(self.R*self.T_inf) ## mass density [kg/m³]
        self.mu_inf  = self.mu_Suth_ref*(self.T_inf/self.T_Suth_ref)**(3/2) * ((self.T_Suth_ref+self.S_Suth)/(self.T_inf+self.S_Suth)) ## [Pa s] | [N s m^-2]
        self.nu_inf  = self.mu_inf / self.rho_inf ## kinematic viscosity [m²/s] --> momentum diffusivity
        
        ## additional
        if ('total_avg_time' in udef_dict):
            self.total_avg_time = udef_dict['total_avg_time']
        if ('total_avg_iter_count' in udef_dict):
            self.total_avg_iter_count = udef_dict['total_avg_iter_count']
        if ('nz' in udef_dict):
            self.nz = udef_dict['nz']
        if ('cp' in udef_dict):
            self.cp = udef_dict['cp']
        if ('cv' in udef_dict):
            self.cv = udef_dict['cv']
        
        # ===
        
        if self.verbose: print(72*'-')
        if self.verbose: even_print('Ma'          , '%0.2f [-]'           % self.Ma          )
        if self.verbose: even_print('Re'          , '%0.1f [-]'           % self.Re          )
        if self.verbose: even_print('Pr'          , '%0.3f [-]'           % self.Pr          )
        if self.verbose: even_print('T_inf'       , '%0.3f [K]'           % self.T_inf       )
        if self.verbose: even_print('p_inf'       , '%0.1f [Pa]'          % self.p_inf       )
        if self.verbose: even_print('kappa'       , '%0.3f [-]'           % self.kappa       )
        if self.verbose: even_print('R'           , '%0.3f [J/(kg·K)]'    % self.R           )
        if self.verbose: even_print('mu_Suth_ref' , '%0.6E [kg/(m·s)]'    % self.mu_Suth_ref )
        if self.verbose: even_print('T_Suth_ref'  , '%0.2f [K]'           % self.T_Suth_ref  )
        if self.verbose: even_print('C_Suth'      , '%0.5e [kg/(m·s·√K)]' % self.C_Suth      )
        if self.verbose: even_print('S_Suth'      , '%0.2f [K]'           % self.S_Suth      )
        if self.verbose: print(72*'-')
        
        self.a_inf     = np.sqrt(self.kappa*self.R*self.T_inf)
        self.U_inf     = self.Ma*self.a_inf
        self.cp        = self.R*self.kappa/(self.kappa-1.)
        self.cv        = self.cp/self.kappa
        self.recov_fac = self.Pr**(1/3)
        self.Taw       = self.T_inf + self.recov_fac*self.U_inf**2/(2*self.cp)
        self.lchar     = self.Re*self.nu_inf/self.U_inf
        
        if self.verbose: even_print('rho_inf'         , '%0.3f [kg/m³]'    % self.rho_inf   )
        if self.verbose: even_print('mu_inf'          , '%0.6E [kg/(m·s)]' % self.mu_inf    )
        if self.verbose: even_print('nu_inf'          , '%0.6E [m²/s]'     % self.nu_inf    )
        if self.verbose: even_print('a_inf'           , '%0.6f [m/s]'      % self.a_inf     )
        if self.verbose: even_print('U_inf'           , '%0.6f [m/s]'      % self.U_inf     )
        if self.verbose: even_print('cp'              , '%0.3f [J/(kg·K)]' % self.cp        )
        if self.verbose: even_print('cv'              , '%0.3f [J/(kg·K)]' % self.cv        )
        #if self.verbose: even_print('recovery factor' , '%0.6f [-]'        % self.recov_fac )
        #if self.verbose: even_print('Taw'             , '%0.3f [K]'        % self.Taw       )
        if self.verbose: even_print('lchar'           , '%0.6E [m]'        % self.lchar     )
        if self.verbose: print(72*'-'+'\n')
        
        # ===
        
        udef_char = [    'Ma',     'Re',     'Pr',     'kappa',    'R',    'p_inf',    'T_inf',    'C_Suth',    'mu_Suth_ref',    'T_Suth_ref' ]
        udef_real = [self.Ma , self.Re , self.Pr , self.kappa, self.R, self.p_inf, self.T_inf, self.C_Suth, self.mu_Suth_ref, self.T_Suth_ref  ]
        udef = dict(zip(udef_char, udef_real))
        
        return udef



import os
import sys
import timeit
from pathlib import PurePosixPath

import numpy as np
import psutil
from mpi4py import MPI
from tqdm import tqdm

from .h5 import h5_chunk_sizer
from .utils import even_print, format_time_string

# ======================================================================

def _calc_mean(self, **kwargs):
    '''
    Calculate mean in [t] --> leaves [x,y,z,1]
    --> save to new RGD file
    -----
    - uses accumulator buffers and does *(1/n) at end to calculate mean
    - allows for low RAM usage, as the time dim can be sub-chunked (ct=N)
    '''
    
    rgd_meta = type(self) ## workaround for using rgd()
    
    if (self.rank==0):
        verbose = True
    else:
        verbose = False
    
    if verbose: print('\n'+'rgd.calc_mean()'+'\n'+72*'-')
    t_start_func = timeit.default_timer()
    
    rx = kwargs.get('rx',1)
    ry = kwargs.get('ry',1)
    rz = kwargs.get('rz',1)
    rt = kwargs.get('rt',1)
    
    fn_rgd_mean  = kwargs.get('fn_rgd_mean',None)
    #sfm         = kwargs.get('scalars',None) ## scalars to take (for mean)
    ti_min       = kwargs.get('ti_min',None)
    favre        = kwargs.get('favre',True)
    reynolds     = kwargs.get('reynolds',True)
    
    ct           = kwargs.get('ct',1) ## number of [t] chunks
    
    force        = kwargs.get('force',False)
    
    chunk_kb         = kwargs.get('chunk_kb',4*1024) ## h5 chunk size: default 4 [MB]
    chunk_constraint = kwargs.get('chunk_constraint',(1,None,None,None)) ## the 'constraint' parameter for sizing h5 chunks
    chunk_base       = kwargs.get('chunk_base',2)
    
    stripe_count   = kwargs.pop('stripe_count'   , 16 ) ## for initializing mean file
    stripe_size_mb = kwargs.pop('stripe_size_mb' , 2  )
    
    if (rt!=1):
        raise AssertionError('rt!=1')
    if (rx*ry*rz != self.n_ranks):
        raise AssertionError('rx*ry*rz != self.n_ranks')
    if (rx>self.nx):
        raise AssertionError('rx>self.nx')
    if (ry>self.ny):
        raise AssertionError('ry>self.ny')
    if (rz>self.nz):
        raise AssertionError('rz>self.nz')
    if (ti_min is not None):
        if not isinstance(ti_min, int):
            raise TypeError('ti_min must be type int')
    
    if self.usingmpi:
        comm4d = self.comm.Create_cart(dims=[rx,ry,rz], periods=[False,False,False], reorder=False)
        t4d = comm4d.Get_coords(self.rank)
        
        rxl_ = np.array_split(np.arange(self.nx,dtype=np.int64),min(rx,self.nx))
        ryl_ = np.array_split(np.arange(self.ny,dtype=np.int64),min(ry,self.ny))
        rzl_ = np.array_split(np.arange(self.nz,dtype=np.int64),min(rz,self.nz))
        #rtl_ = np.array_split(np.arange(self.nt,dtype=np.int64),min(rt,self.nt))
        
        rxl = [[b[0],b[-1]+1] for b in rxl_ ]
        ryl = [[b[0],b[-1]+1] for b in ryl_ ]
        rzl = [[b[0],b[-1]+1] for b in rzl_ ]
        #rtl = [[b[0],b[-1]+1] for b in rtl_ ]
        
        rx1, rx2 = rxl[t4d[0]]; nxr = rx2 - rx1
        ry1, ry2 = ryl[t4d[1]]; nyr = ry2 - ry1
        rz1, rz2 = rzl[t4d[2]]; nzr = rz2 - rz1
        #rt1, rt2 = rtl[t4d[3]]; ntr = rt2 - rt1
    else:
        nxr = self.nx
        nyr = self.ny
        nzr = self.nz
        #ntr = self.nt
    
    nx = self.nx
    ny = self.ny
    nz = self.nz
    nt = self.nt
    
    ## mean file name (for writing)
    if (fn_rgd_mean is None):
        fname_path = os.path.dirname(self.fname)
        fname_base = os.path.basename(self.fname)
        fname_root, fname_ext = os.path.splitext(fname_base)
        fname_mean_h5_base = fname_root+'_mean.h5'
        #fn_rgd_mean = os.path.join(fname_path, fname_mean_h5_base)
        fn_rgd_mean = str(PurePosixPath(fname_path, fname_mean_h5_base))
        #fn_rgd_mean = Path(fname_path, fname_mean_h5_base)
    
    if verbose: even_print('fn_rgd'          , self.fname    )
    if verbose: even_print('fn_rgd_mean'     , fn_rgd_mean   )
    if verbose: even_print('do Favre avg'    , str(favre)    )
    if verbose: even_print('do Reynolds avg' , str(reynolds) )
    if verbose: print(72*'-')
    if verbose: even_print('nx',f'{self.nx:d}')
    if verbose: even_print('ny',f'{self.ny:d}')
    if verbose: even_print('nz',f'{self.nz:d}')
    if verbose: even_print('nt',f'{self.nt:d}')
    if verbose: print(72*'-')
    if verbose: even_print('rx',f'{rx:d}')
    if verbose: even_print('ry',f'{ry:d}')
    if verbose: even_print('rz',f'{rz:d}')
    if verbose: even_print('ct',f'{ct:d}')
    if verbose: print(72*'-')
    
    ## get times to take for avg
    if (ti_min is not None):
        ti_for_avg = np.copy( self.ti[ti_min:] )
    else:
        ti_for_avg = np.copy( self.ti )
    
    nt_avg       = ti_for_avg.shape[0]
    t_avg_start  = self.t[ti_for_avg[0]]
    t_avg_end    = self.t[ti_for_avg[-1]]
    duration_avg = t_avg_end - t_avg_start
    
    #if not isinstance(ct, (int,np.int32,np.int64)):
    if not isinstance(ct, int):
        raise ValueError
    if (ct<1):
        raise ValueError
    
    ## [t] sub chunk range
    ctl_ = np.array_split( ti_for_avg, min(ct,nt_avg) )
    ctl = [[b[0],b[-1]+1] for b in ctl_ ]
    
    ## check that no sub ranges are <=1
    for a_ in [ ctl_[1]-ctl_[0] for ctl_ in ctl ]:
        if (a_ <= 1):
            raise ValueError
    
    ## assert constant Δt, later attach dt as attribute to mean file
    dt0 = np.diff(self.t)[0]
    if not np.all(np.isclose(np.diff(self.t), dt0, rtol=1e-7)):
        raise ValueError
    
    if verbose: even_print('n timesteps avg','%i/%i'%(nt_avg,self.nt))
    if verbose: even_print('t index avg start','%i'%(ti_for_avg[0],))
    if verbose: even_print('t index avg end','%i'%(ti_for_avg[-1],))
    if verbose: even_print('t avg start','%0.2f [-]'%(t_avg_start,))
    if verbose: even_print('t avg end','%0.2f [-]'%(t_avg_end,))
    if verbose: even_print('duration avg','%0.2f [-]'%(duration_avg,))
    if verbose: even_print('Δt','%0.2f [-]'%(dt0,))
    #if verbose: print(72*'-')
    
    ## performance
    t_read = 0.
    t_write = 0.
    data_gb_read = 0.
    data_gb_write = 0.
    
    #scalars_re = ['u','v','w','p','T','rho']
    scalars_fv = ['u','v','w','T'] ## 'p','rho'
    
    ## do a loop through to get names
    scalars_mean_names  = []
    #scalars_mean_dtypes = []
    for scalar in self.scalars:
        
        ##dtype = self.scalars_dtypes_dict[scalar]
        #dtype = np.float64 ## always save mean as double
        
        if reynolds:
            if True: ## always
                sc_name = scalar
                scalars_mean_names.append(sc_name)
                #scalars_mean_dtypes.append(dtype)
        if favre:
            if (scalar in scalars_fv):
                sc_name = f'r_{scalar}'
                scalars_mean_names.append(sc_name)
                #scalars_mean_dtypes.append(dtype)
    
    #with rgd(fn_rgd_mean, 'w', force=force, driver='mpio', comm=MPI.COMM_WORLD) as hf_mean:
    with rgd_meta(fn_rgd_mean, 'w', force=force, driver=self.driver, comm=self.comm, stripe_count=stripe_count, stripe_size_mb=stripe_size_mb) as hf_mean:
        
        ## initialize the mean file from the opened unsteady rgd file
        hf_mean.init_from_rgd(self.fname)
        
        ## set some top-level attributes (in MEAN file)
        hf_mean.attrs['duration_avg'] = duration_avg ## duration of mean
        #hf_mean.attrs['duration_avg'] = self.duration
        hf_mean.attrs['dt'] = dt0
        #hf_mean.attrs['fclass'] = 'rgd'
        hf_mean.attrs['fsubtype'] = 'mean'
        
        if verbose: print(72*'-')
        
        # === initialize datasets in mean file
        for scalar in self.scalars:
            
            data_gb_mean = np.dtype(np.float64).itemsize * self.nx*self.ny*self.nz * 1 / 1024**3
            
            shape  = (1,self.nz,self.ny,self.nx)
            chunks = h5_chunk_sizer(nxi=shape, constraint=chunk_constraint, size_kb=chunk_kb, base=chunk_base, itemsize=np.dtype(np.float64).itemsize)
            
            if reynolds:
                
                ## do the Re mean of all scalars in file, regardless whether explicitly in scalars_re or not
                #if scalar in scalars_re:
                if True:
                    
                    if ('data/%s'%scalar in hf_mean):
                        del hf_mean['data/%s'%scalar]
                    if verbose:
                        even_print( f'initializing data/{scalar}' , f'{data_gb_mean:0.3f} [GB]' )
                    dset = hf_mean.create_dataset(
                                                f'data/{scalar}',
                                                shape=shape,
                                                dtype=np.float64, ## mean dsets always double
                                                chunks=chunks,
                                                )
                    hf_mean.scalars.append('data/%s'%scalar)
                    
                    chunk_kb_ = np.prod(dset.chunks)*dset.dtype.itemsize / 1024. ## actual
                    if verbose:
                        even_print('chunk shape (t,z,y,x)','%s'%str(dset.chunks))
                        even_print('chunk size','%i [KB]'%int(round(chunk_kb_)))
            
            if favre:
                
                if (scalar in scalars_fv):
                    if ('data/%s_fv'%scalar in hf_mean):
                        del hf_mean['data/%s_fv'%scalar]
                    if verbose:
                        even_print( f'initializing data/{scalar}_fv' , f'{data_gb_mean:0.3f} [GB]' )
                    dset = hf_mean.create_dataset(f'data/{scalar}_fv',
                                                  shape=shape,
                                                  dtype=np.float64, ## mean dsets always double
                                                  chunks=chunks,
                                                  )
                    hf_mean.scalars.append('data/%s_fv'%scalar)
                    
                    chunk_kb_ = np.prod(dset.chunks)*dset.dtype.itemsize / 1024. ## actual
                    if verbose:
                        even_print('chunk shape (t,z,y,x)','%s'%str(dset.chunks))
                        even_print('chunk size','%i [KB]'%int(round(chunk_kb_)))
        
        if self.usingmpi: self.comm.Barrier()
        #if verbose: print(72*'-')
        
        ## accumulator array for local rank --> initialize
        data_sum = np.zeros(shape=(nxr,nyr,nzr,1), dtype={'names':scalars_mean_names, 'formats':[np.float64 for _ in scalars_mean_names]})
        
        # ==========================================================
        # check memory
        # ==========================================================
        
        hostname = MPI.Get_processor_name()
        mem_free_gb = psutil.virtual_memory().free / 1024**3
        G = self.comm.gather([ self.rank , hostname , mem_free_gb ], root=0)
        G = self.comm.bcast(G, root=0)
        
        host_mem = {}
        for rank, host, mem in G:
            if host not in host_mem or mem < host_mem[host]:
                host_mem[host] = mem
        total_free = sum(host_mem.values())
        
        if verbose:
            print(72*'-')
            for key,value in host_mem.items():
                even_print(f'RAM free {key}', f'{int(np.floor(value)):d} [GB]')
            even_print('RAM free (local,min)', f'{int(np.floor(min(host_mem.values()))):d} [GB]')
            even_print('RAM free (global)', f'{int(np.floor(total_free)):d} [GB]')
        
        shape_read_local = (nxr,nyr,nzr,nt//ct)
        data_gb_local    = np.dtype(np.float64).itemsize * np.prod(shape_read_local) / 1024**3
        if verbose: even_print('read shape (local)', f'[{nxr:d},{nyr:d},{nzr:d},{nt//ct:d}]')
        if verbose: even_print('read size (local)', f'{int(np.ceil(data_gb_local)):d} [GB]')
        
        shape_read_global = (nx,ny,nz,nt//ct)
        data_gb_global    = np.dtype(np.float64).itemsize * np.prod(shape_read_global) / 1024**3
        if verbose: even_print('read shape (global)'   , f'[{nx:d},{ny:d},{nz:d},{nt//ct:d}]')
        if verbose: even_print('read size (global)'    , f'{int(np.ceil(data_gb_global)):d} [GB]')
        if verbose: even_print('read size (global) ×3' , f'{int(np.ceil(data_gb_global*3)):d} [GB]')
        ram_usage_est = data_gb_global*3/total_free
        
        if verbose: even_print('RAM usage estimate', f'{100*ram_usage_est:0.1f} [%]')
        self.comm.Barrier()
        if (ram_usage_est>0.60):
            print('RAM consumption might be too high. exiting.')
            self.comm.Abort(1)
        
        # ==========================================================
        # main loop
        # ==========================================================
        
        if self.usingmpi: self.comm.Barrier()
        if verbose: print(72*'-')
        
        if verbose:
            progress_bar = tqdm(
                total=ct,
                ncols=100,
                desc='mean',
                leave=True,
                file=sys.stdout,
                mininterval=0.1,
                smoothing=0.,
                #bar_format="\033[B{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}\033[A\n\b",
                bar_format="{l_bar}{bar}| {n}/{total} [{percentage:.1f}%] {elapsed}/{remaining}",
                ascii="░█",
                colour='#FF6600',
                )
        
        ct_counter=0
        for ctl_ in ctl:
            ct_counter += 1
            ct1, ct2 = ctl_
            ntc = ct2 - ct1
            
            if (ct>1):
                if verbose:
                    mesg = f'[t] sub chunk {ct_counter:d}/{ct:d}'
                    tqdm.write( mesg )
                    tqdm.write( '-'*len(mesg) )
            
            ## Read ρ for Favre averaging
            if favre:
                
                dset = self['data/rho']
                
                if self.usingmpi: self.comm.Barrier()
                t_start = timeit.default_timer()
                
                if self.usingmpi: 
                    with dset.collective:
                        rho = dset[ct1:ct2,rz1:rz2,ry1:ry2,rx1:rx2].T
                else:
                    rho = dset[ct1:ct2,:,:,:].T
                
                if self.usingmpi: self.comm.Barrier()
                t_delta = timeit.default_timer() - t_start
                
                data_gb = dset.dtype.itemsize * self.nx*self.ny*self.nz * ntc / 1024**3
                if verbose:
                    txt = even_print('read: rho', '%0.2f [GB]  %0.2f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True)
                    tqdm.write(txt)
                
                t_read       += t_delta
                data_gb_read += data_gb
                
                ## convert to double
                rho = rho.astype(np.float64)
            
            ## Read data, perform sum Σ
            for scalar in self.scalars:
                
                dset = self[f'data/{scalar}']
                
                if self.usingmpi: self.comm.Barrier()
                t_start = timeit.default_timer()
                
                if self.usingmpi:
                    with dset.collective:
                        data = dset[ct1:ct2,rz1:rz2,ry1:ry2,rx1:rx2].T
                else:
                    data = dset[ct1:ct2,:,:,:].T
                
                if self.usingmpi: self.comm.Barrier()
                t_delta = timeit.default_timer() - t_start
                
                data_gb = dset.dtype.itemsize * self.nx*self.ny*self.nz * ntc / 1024**3
                
                if verbose:
                    txt = even_print('read: %s'%scalar, '%0.2f [GB]  %0.2f [s]  %0.3f [GB/s]'%(data_gb,t_delta,(data_gb/t_delta)), s=True)
                    tqdm.write(txt)
                
                t_read       += t_delta
                data_gb_read += data_gb
                
                ## convert to double
                data = data.astype(np.float64)
                
                # === do sum, add to accumulator
                if reynolds:
                    sc_name = scalar
                    data_sum[sc_name] += np.sum(data, axis=-1, dtype=np.float64, keepdims=True)
                if favre:
                    if (scalar in scalars_fv):
                        sc_name = f'r_{scalar}'
                        data_sum[sc_name] += np.sum(data*rho, axis=-1, dtype=np.float64, keepdims=True)
                
                if self.usingmpi: self.comm.Barrier()
            
            ## check RAM
            #mem_avail_gb = psutil.virtual_memory().available/1024**3
            mem_free_gb  = psutil.virtual_memory().free/1024**3
            if verbose:
                tqdm.write(even_print('mem free', '%0.1f [GB]'%mem_free_gb, s=True))
            
            if verbose: progress_bar.update()
            if verbose: tqdm.write(72*'-')
        if verbose: progress_bar.close()
        
        # ==========================================================
        # multiply accumulators by (1/n)
        # ==========================================================
        
        for scalar in self.scalars:
            if reynolds:
                sc_name = scalar
                data_sum[sc_name] *= (1/nt_avg)
            if favre:
                if (scalar in scalars_fv):
                    sc_name = f'r_{scalar}'
                    data_sum[sc_name] *= (1/nt_avg)
        
        # ==========================================================
        # 'data_sum' now contains averages, not sums!
        # ==========================================================
        
        ## Favre avg : φ_tilde = avg[ρ·φ]/avg[ρ]
        rho_mean = np.copy( data_sum['rho'] )
        
        # === write
        for scalar in self.scalars:
            
            if reynolds:
                
                dset = hf_mean[f'data/{scalar}']
                
                data_out = np.copy( data_sum[scalar] )
                
                ## if storing as single precision, pre-convert
                if (dset.dtype==np.float32):
                    data_out = np.copy( data_sum[scalar].astype(np.float32) )
                
                if self.usingmpi: self.comm.Barrier()
                t_start = timeit.default_timer()
                if self.usingmpi:
                    with dset.collective:
                        dset[:,rz1:rz2,ry1:ry2,rx1:rx2] = data_out.T
                else:
                    dset[:,:,:,:] = data_out.T
                if self.usingmpi: self.comm.Barrier()
                t_delta = timeit.default_timer() - t_start
                
                data_gb_mean = data_out.dtype.itemsize * self.nx*self.ny*self.nz * 1 / 1024**3
                
                if verbose:
                    txt = even_print(f'write: {scalar}', '%0.2f [GB]  %0.2f [s]  %0.3f [GB/s]'%(data_gb_mean,t_delta,(data_gb_mean/t_delta)), s=True)
                    tqdm.write(txt)
                
                t_write       += t_delta
                data_gb_write += data_gb_mean
            
            if favre:
                if (scalar in scalars_fv):
                    
                    dset = hf_mean[f'data/{scalar}_fv']
                    
                    ## φ_tilde = avg[ρ·φ]/avg[ρ]
                    data_out = np.copy( data_sum[f'r_{scalar}'] / rho_mean )
                    
                    ## if storing as single precision, pre-convert
                    if (dset.dtype==np.float32):
                        data_out = np.copy( data_out.astype(np.float32) )
                    
                    if self.usingmpi: self.comm.Barrier()
                    t_start = timeit.default_timer()
                    if self.usingmpi:
                        with dset.collective:
                            dset[:,rz1:rz2,ry1:ry2,rx1:rx2] = data_out.T
                    else:
                        dset[:,:,:,:] = data_out.T
                    if self.usingmpi: self.comm.Barrier()
                    t_delta = timeit.default_timer() - t_start
                    
                    data_gb_mean = data_out.dtype.itemsize * self.nx*self.ny*self.nz * 1 / 1024**3
                    
                    if verbose:
                        txt = even_print(f'write: {scalar}_fv', '%0.2f [GB]  %0.2f [s]  %0.3f [GB/s]'%(data_gb_mean,t_delta,(data_gb_mean/t_delta)), s=True)
                        tqdm.write(txt)
                    
                    t_write       += t_delta
                    data_gb_write += data_gb_mean
        
        # === replace dims/t array --> take last time of series
        t = np.array([self.t[-1]],dtype=np.float64)
        if ('dims/t' in hf_mean):
            del hf_mean['dims/t']
        hf_mean.create_dataset('dims/t', data=t)
        
        if hasattr(hf_mean, 'duration_avg'):
            if verbose: even_print('duration avg', '%0.2f [-]'%hf_mean.duration_avg)
    
    if verbose: print(72*'-')
    if verbose: even_print('time read',format_time_string(t_read))
    if verbose: even_print('time write',format_time_string(t_write))
    if verbose: even_print('read total avg', '%0.2f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb_read,t_read,(data_gb_read/t_read)))
    if verbose: even_print('write total avg', '%0.2f [GB]  %0.3f [s]  %0.3f [GB/s]'%(data_gb_write,t_write,(data_gb_write/t_write)))
    
    ## report file
    self.comm.Barrier()
    if verbose:
        print(72*'-')
        even_print( os.path.basename(fn_rgd_mean), f'{(os.path.getsize(fn_rgd_mean)/1024**3):0.1f} [GB]')
    if verbose: print(72*'-')
    if verbose: print('total time : rgd.calc_mean() : %s'%format_time_string((timeit.default_timer() - t_start_func)))
    if verbose: print(72*'-')
    return



import numpy as np

'''
========================================================================
Utilities
========================================================================
'''

# ======================================================================

def even_print(label, output, **kwargs):
    '''
    print/return a fixed width message
    '''
    terminal_width = kwargs.get('terminal_width',72)
    s              = kwargs.get('s',False) ## return string
    
    ndots = (terminal_width-2) - len(label) - len(output)
    text = label+' '+ndots*'.'+' '+output
    if s:
        return text
    else:
        #sys.stdout.write(text)
        print(text)
        return

def format_time_string(tsec):
    '''
    format seconds as dd:hh:mm:ss
    '''
    m, s = divmod(abs(int(round(tsec))),60)
    h, m = divmod(m,60)
    d, h = divmod(h,24)
    d    = int(round(d))
    h    = int(round(h))
    m    = int(round(m))
    s    = int(round(s))
    if (d==0) and (h==0) and (m==0):
        s_out = f'{s:d}s'
    elif (d==0) and (h==0):
        s_out = f'{m:d}m:{s:02d}s'
    elif (d==0):
        s_out = f'{h:d}h:{m:02d}m:{s:02d}s'
    else:
        s_out = f'{d:d}d:{h:02d}h:{m:02d}m:{s:02d}s'
    if (tsec >= 0):
        return s_out
    else:
        return f'-{s_out}'

def format_nbytes(size):
    '''
    format a number of bytes to [B],[KB],[MB],[GB],[TB]
    '''
    if not isinstance(size,(int,float)):
        raise ValueError('arg should be of type int or float')
    if (size<1024):
        size_fmt, size_unit = size, '[B]'
    elif (size>1024) and (size<=1024**2):
        size_fmt, size_unit = size/1024, '[KB]'
    elif (size>1024**2) and (size<=1024**3):
        size_fmt, size_unit = size/1024**2, '[MB]'
    elif (size>1024**3) and (size<=1024**4):
        size_fmt, size_unit = size/1024**3, '[GB]'
    else:
        size_fmt, size_unit = size/1024**4, '[TB]'
    return size_fmt, size_unit

def step(a,b,x,order=3):
    '''
    Polynomial step function
    https://en.wikipedia.org/wiki/Smoothstep
    '''
    x = (x-a)/(b-a)
    if (order==1):
        y = -2*x**3 + 3*x**2
    elif (order==2):
        y = 6*x**5 - 15*x**4 + 10*x**3
    elif (order==3):
        y = -20*x**7 + 70*x**6 - 84*x**5 + 35*x**4
    else:
        raise NotImplementedError
    y = np.clip(y,0.,1.)
    return y
